from time import monotonic, sleep

class IntervalTimer(object):
    """ this is a timer that does not drift (but is has jitter). """
    VERBOSE = 0                    
    def __init__(self, periodInSeconds, activeWaitLimit=0.0):
        self._period = periodInSeconds
        self._activeWaitLimit = activeWaitLimit
        self._nextIntervalStart = monotonic() + self._period

    
    def setPeriod(self, period):
        self._period = period

        
    def reset(self):
        self._nextIntervalStart = monotonic() + self._period

    
    def waitForNextIntervalBegin(self):
        """
        :param return: 0.0 if timer is in scheduled plan, otherwise seconds how far timer is behind schedule
        """
        behindSchedule = 0.0
        now = monotonic()
        dt = self._nextIntervalStart - now
        if dt <= 0:
            behindSchedule = abs(dt)
        elif dt > self._activeWaitLimit:
            # normal sleep
            sleep(dt)
        else:
            # active wait, time is too short for a sleep call
            while dt > 0:
                dt = self._nextIntervalStart - monotonic()
        self._nextIntervalStart += self._period 
        return behindSchedule
    sleep = waitForNextIntervalBegin  # alias
    



#!/usr/bin/env python
from lxml.etree import ETCompatXMLParser, QName, Element, SubElement, tostring, fromstring

import random
import socket
import struct
import time
import uuid
import threading
import platform
import selectors
import re
from collections import deque
import traceback
import logging
import urllib
from urllib.parse import urlparse
import queue
from dataclasses import dataclass, field
from typing import Any

from .netconn import get_ipv4_addresses, get_ip_for_adapter, get_ipv4_ips

try:
    from sdc11073.commlog import getCommunicationLogger
except ImportError:
    class NullLogger(object):
        """ This is a dummy logger that does nothing."""

        def __getattr__(self, name):
            return self.do_nothing

        def do_nothing(self, *args, **kwargs):
            pass


    communicationLogger = NullLogger()


    def getCommunicationLogger():
        return communicationLogger

BUFFER_SIZE = 0xffff
APP_MAX_DELAY = 500  # miliseconds
DP_MAX_TIMEOUT = 5000  # 5 seconds

_NETWORK_ADDRESSES_CHECK_TIMEOUT = 5

MULTICAST_PORT = 3702
MULTICAST_IPV4_ADDRESS = "239.255.255.250"
MULTICAST_OUT_TTL = 15  # Time To Live for multicast_out

UNICAST_UDP_REPEAT = 2
UNICAST_UDP_MIN_DELAY = 50
UNICAST_UDP_MAX_DELAY = 250
UNICAST_UDP_UPPER_DELAY = 500

MULTICAST_UDP_REPEAT = 4
MULTICAST_UDP_MIN_DELAY = 50
MULTICAST_UDP_MAX_DELAY = 250
MULTICAST_UDP_UPPER_DELAY = 500

# pylint: disable=protected-access, redefined-outer-name, len-as-condition, attribute-defined-outside-init

# values  acc to
#  http://docs.oasis-open.org/ws-dd/discovery/1.1/wsdd-discovery-1.1-spec.html
NS_A = "http://www.w3.org/2005/08/addressing"  # ws-addressing
NS_D = 'http://docs.oasis-open.org/ws-dd/ns/discovery/2009/01'  # ws-discovery
NS_S = "http://www.w3.org/2003/05/soap-envelope"  # "http://www.w3.org/2003/05/soap-envelope"
NS_DPWS = 'http://docs.oasis-open.org/ws-dd/ns/dpws/2009/01'
ACTION_HELLO = NS_D + '/Hello'  # "http://docs.oasis-open.org/ws-dd/ns/discovery/2009/01/Hello"
ACTION_BYE = NS_D + '/Bye'  # "http://docs.oasis-open.org/ws-dd/ns/discovery/2009/01/Bye"
ACTION_PROBE = NS_D + '/Probe'  # "http://docs.oasis-open.org/ws-dd/ns/discovery/2009/01/Probe"
ACTION_PROBE_MATCH = NS_D + '/ProbeMatches'  # "http://docs.oasis-open.org/ws-dd/ns/discovery/2009/01/ProbeMatches"
ACTION_RESOLVE = NS_D + '/Resolve'  # "http://docs.oasis-open.org/ws-dd/ns/discovery/2009/01/Resolve"
ACTION_RESOLVE_MATCH = NS_D + '/ResolveMatches'  # "http://docs.oasis-open.org/ws-dd/ns/discovery/2009/01/ResolveMatches"

ADDRESS_ALL = "urn:docs-oasis-open-org:ws-dd:ns:discovery:2009:01"  # format acc to RFC 2141
WSA_ANONYMOUS = NS_A + '/anonymous'
MATCH_BY_LDAP = NS_D + '/ldap'  # "http://docs.oasis-open.org/ws-dd/ns/discovery/2009/01/ldap"
MATCH_BY_URI = NS_D + '/rfc3986'  # "http://docs.oasis-open.org/ws-dd/ns/discovery/2009/01/rfc3986"
MATCH_BY_UUID = NS_D + '/uuid'  # "http://docs.oasis-open.org/ws-dd/ns/discovery/2009/01/uuid"
MATCH_BY_STRCMP = NS_D + '/strcmp0'  # "http://docs.oasis-open.org/ws-dd/ns/discovery/2009/01/strcmp0"

_IP_BLACKLIST = ('0.0.0.0', None)  # None can happen if an adapter does not have any IP address

# these time constants control the send loop
SEND_LOOP_IDLE_SLEEP = 0.1
SEND_LOOP_BUSY_SLEEP = 0.01


class WsaTag(QName):
    def __init__(self, localname):
        super().__init__(NS_A, localname)


class WsdTag(QName):
    def __init__(self, localname):
        super().__init__(NS_D, localname)


class S12Tag(QName):
    def __init__(self, localname):
        super().__init__(NS_S, localname)


_namespaces_map = {'wsd': NS_D, 'wsa': NS_A, 's12': NS_S, 'dpws': NS_DPWS}


def types_info(types):
    # helper for logging
    return [str(t) for t in types] if types else types


def _getPrefix(nsmap, ns):
    for prefix, namespace in nsmap.items():
        if namespace == ns:
            return prefix


class URI:

    def __init__(self, uri):
        i1 = uri.find(":")
        i2 = uri.find("@")
        self._scheme = uri[:i1]
        if i2 != -1:
            self._authority = uri[i1 + 1: i2]
            self._path = uri[i2 + 1:]
        else:
            self._authority = ""
            self._path = uri[i1 + 1:]

    def getScheme(self):
        return self._scheme

    def getAuthority(self):
        return self._authority

    def getPath(self):
        return self._path

    def getPathExQueryFragment(self):
        i = self._path.find("?")
        path = self.getPath()
        if i != -1:
            return path[:self._path.find("?")]
        else:
            return path


class Scope:

    def __init__(self, value, matchBy=None):
        self._matchBy = matchBy
        self._value = value

    def getMatchBy(self):
        return self._matchBy

    def getValue(self):
        return self._value

    def getQuotedValue(self):
        return self._value.replace(' ', '%20')

    def __repr__(self):
        if self.getMatchBy() is None or len(self.getMatchBy()) == 0:
            return self.getValue()
        else:
            return self.getMatchBy() + ":" + self.getValue()


class ProbeResolveMatch:

    def __init__(self, epr, types, scopes, xAddrs, metadataVersion):
        self._epr = epr
        self._types = types
        self._scopes = scopes
        self._xAddrs = xAddrs
        self._metadataVersion = metadataVersion

    def getEPR(self):
        return self._epr

    def getTypes(self):
        return self._types

    def getScopes(self):
        return self._scopes

    def getXAddrs(self):
        return self._xAddrs

    def getMetadataVersion(self):
        return self._metadataVersion

    def __repr__(self):
        return "ProbeResolveMatch(EPR:%s Types:%s Scopes:%s XAddrs:%s Metadata Version:%s)" % \
               (self.getEPR(), types_info(self.getTypes()),
                [str(s) for s in self.getScopes()],
                self.getXAddrs(),
                self.getMetadataVersion())


class SoapEnvelope:

    def __init__(self, messageID=None):
        self._action = ""
        self._messageId = messageID or uuid.uuid4().urn
        self._relatesTo = ""
        self._relationshipType = None
        self._to = ""
        self._replyTo = ""
        self._instanceId = ""
        self._sequenceId = ""
        self._messageNumber = ""
        self._epr = ""
        self._types = []
        self._scopes = []
        self._xAddrs = []
        self._metadataVersion = "1"
        self._probeResolveMatches = []

    def getAction(self):
        return self._action

    def setAction(self, action):
        self._action = action

    def getMessageId(self):
        return self._messageId

    def getRelatesTo(self):
        return self._relatesTo

    def setRelatesTo(self, relatesTo):
        self._relatesTo = relatesTo

    def getRelationshipType(self):
        return self._relationshipType

    def setRelationshipType(self, relationshipType):
        self._relationshipType = relationshipType

    def getTo(self):
        return self._to

    def setTo(self, to):
        self._to = to

    def getReplyTo(self):
        return self._replyTo

    def setReplyTo(self, replyTo):
        self._replyTo = replyTo

    def getInstanceId(self):
        return self._instanceId

    def setInstanceId(self, instanceId):
        self._instanceId = instanceId

    def getSequenceId(self):
        return self._sequenceId

    def setSequenceId(self, sequenceId):
        self._sequenceId = sequenceId

    def getEPR(self):
        return self._epr

    def setEPR(self, epr):
        self._epr = epr

    def getMessageNumber(self):
        return self._messageNumber

    def setMessageNumber(self, messageNumber):
        self._messageNumber = messageNumber

    def getTypes(self):
        return self._types

    def setTypes(self, types):
        self._types = types

    def getScopes(self):
        return self._scopes

    def setScopes(self, scopes):
        self._scopes = scopes

    def getXAddrs(self):
        return self._xAddrs

    def setXAddrs(self, xAddrs):
        self._xAddrs = xAddrs

    def getMetadataVersion(self):
        return self._metadataVersion

    def setMetadataVersion(self, metadataVersion):
        self._metadataVersion = metadataVersion

    def getProbeResolveMatches(self):
        return self._probeResolveMatches

    def setProbeResolveMatches(self, probeResolveMatches):
        self._probeResolveMatches = probeResolveMatches


def matchScope(src, target, matchBy):
    """ This implementation correctly handles "%2F" (== '/') encoded values"""
    if matchBy == "" or matchBy is None or matchBy == MATCH_BY_LDAP or matchBy == MATCH_BY_URI or matchBy == MATCH_BY_UUID:
        src = urllib.parse.urlsplit(src)
        target = urllib.parse.urlsplit(target)
        if src.scheme.lower() != target.scheme.lower():
            return False
        if src.netloc.lower() != target.netloc.lower():
            return False
        if src.path == target.path:
            return True
        srcPathElements = src.path.split('/')
        targetPathElements = target.path.split('/')
        srcPathElements = [urllib.parse.unquote(elem) for elem in srcPathElements]
        targetPathElements = [urllib.parse.unquote(elem) for elem in targetPathElements]
        if len(srcPathElements) > len(targetPathElements):
            return False
        for i, elem in enumerate(srcPathElements):
            if targetPathElements[i] != elem:
                return False
        return True
    elif matchBy == MATCH_BY_STRCMP:
        return src == target
    else:
        return False


def matchType(type1, type2):
    return type1.namespace == type2.namespace and type1.localname == type2.localname


_ascii_letters = 'abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ'


def getRandomStr():
    return "".join([random.choice(_ascii_letters) for x in range(10)])


def _parseTypes(parentNode):
    types = []
    typesNode = parentNode.find('wsd:Types', _namespaces_map)
    if typesNode is not None:
        _types = [] if not typesNode.text else typesNode.text.split()
        for t in _types:
            elements = t.split(':')
            prefix = None if len(elements) == 1 else elements[0]  # None means default namespace
            localname = elements[-1]
            q = QName(typesNode.nsmap[prefix], localname)
            types.append(q)
    return types


def _parseScopes(parentNode):
    scopesNode = parentNode.find('wsd:Scopes', _namespaces_map)
    if scopesNode is not None:
        matchBy = scopesNode.attrib.get("MatchBy")
        scopes = [] if not scopesNode.text else scopesNode.text.split()
        return [Scope(item, matchBy) for item in scopes]
    else:
        return []


def _parseXAddrs(parentNode):
    xAddrNode = parentNode.find('wsd:XAddrs', _namespaces_map)
    if xAddrNode is not None:
        return [] if not xAddrNode.text else xAddrNode.text.split()
    else:
        return []


def _parseEpr(parentNode):
    """

    :param parentNode: direct parent of wsa:EndpointReference node
    :return: epr address
    """
    eprNode = parentNode.find('wsa:EndpointReference', _namespaces_map)
    if eprNode is not None:
        addressNode = eprNode.find('wsa:Address', _namespaces_map)
        return addressNode.text
    return ''


def _parseMetaDataVersion(parentNode):
    mdvNode = parentNode.find('wsd:MetadataVersion', _namespaces_map)
    if mdvNode is not None:
        return mdvNode.text
    return ''


def _parseAppSequence(headerNode, env):
    appSeqNode = headerNode.find('wsd:AppSequence', _namespaces_map)
    if appSeqNode is not None:
        env.setInstanceId(appSeqNode.attrib.get("InstanceId"))
        env.setSequenceId(appSeqNode.attrib.get("SequenceId"))
        env.setMessageNumber(appSeqNode.attrib.get("MessageNumber"))


def _parseRelatesTo(headerNode, env):
    relatesTo = headerNode.find('wsa:RelatesTo', _namespaces_map)
    if relatesTo is not None:
        env.setRelatesTo(relatesTo.text)
        rel_type = relatesTo.attrib.get('RelationshipType')
        if rel_type:
            env.setRelationshipType(rel_type)


def _parseReplyTo(headerNode, env):
    replyTo = headerNode.find('wsa:ReplyTo', _namespaces_map)
    if replyTo is not None:
        env.setReplyTo(replyTo.text)


def parseEnvelope(data, ipAddr, logger):
    parser = ETCompatXMLParser()
    try:
        dom = fromstring(data, parser=parser)
    except Exception as ex:
        logger.error('load error "%s" in "%s"', ex, data)
        return

    header = dom.find('s12:Header', _namespaces_map)
    body = dom.find('s12:Body', _namespaces_map)
    if header is None or body is None:
        logger.error('received message from {} is not a soap message: {}', ipAddr, data)
        return None

    msgNode = body[0]

    msgId = header.find('wsa:MessageID', _namespaces_map)
    msgId = None if msgId is None else msgId.text
    soapAction = header.find('wsa:Action', _namespaces_map)
    if soapAction is None:
        # this is something else, ignore it
        return

    soapAction = soapAction.text

    env = SoapEnvelope(msgId)
    env.setAction(soapAction)

    to = header.find('wsa:To', _namespaces_map)
    if to is not None:
        env.setTo(to.text)

    # parse action specific data
    try:
        if soapAction == ACTION_PROBE:
            _parseReplyTo(header, env)
            env.getTypes().extend(_parseTypes(msgNode))
            env.getScopes().extend(_parseScopes(msgNode))
            return env
        elif soapAction == ACTION_PROBE_MATCH:
            _parseRelatesTo(header, env)
            _parseAppSequence(header, env)
            pmNodes = msgNode.findall('wsd:ProbeMatch', _namespaces_map)
            for node in pmNodes:
                epr = _parseEpr(node)
                types = _parseTypes(node)
                scopes = _parseScopes(node)
                xAddrs = _parseXAddrs(node)
                mdv = _parseMetaDataVersion(node)
                env.getProbeResolveMatches().append(ProbeResolveMatch(epr, types, scopes, xAddrs, mdv))
            return env
        elif soapAction == ACTION_RESOLVE:
            _parseReplyTo(header, env)
            env.setEPR(_parseEpr(msgNode))
            return env
        elif soapAction == ACTION_RESOLVE_MATCH:
            _parseRelatesTo(header, env)
            _parseAppSequence(header, env)
            resolveMatchNode = msgNode.find('wsd:ResolveMatch', _namespaces_map)
            if resolveMatchNode is not None:
                epr = _parseEpr(resolveMatchNode)
                types = _parseTypes(resolveMatchNode)
                scopes = _parseScopes(resolveMatchNode)
                xAddrs = _parseXAddrs(resolveMatchNode)
                mdv = _parseMetaDataVersion(resolveMatchNode)
                env.setProbeResolveMatches([ProbeResolveMatch(epr, types, scopes, xAddrs, mdv)])
            return env
        elif soapAction == ACTION_BYE:
            _parseAppSequence(header, env)
            env.setEPR(_parseEpr(msgNode))
            return env
        elif soapAction == ACTION_HELLO:
            _parseAppSequence(header, env)
            env.setEPR(_parseEpr(msgNode))
            env.getTypes().extend(_parseTypes(msgNode))
            env.getScopes().extend(_parseScopes(msgNode))
            env.setXAddrs(_parseXAddrs(msgNode))
            env.setMetadataVersion(_parseMetaDataVersion(msgNode))
            return env
    except:
        logger.error('Parse Error %s:', traceback.format_exc())
        logger.error('parsed data is from %r, data: %r:', ipAddr, data)
        return


class MessageCreator:
    """This class provides methods to create messages from SoapEnvelope instances."""
    def createMessage(self, env) -> bytes:
        action = env.getAction()
        if action == ACTION_PROBE:
            doc, header, body = self.createProbeMessage(env)
        elif action == ACTION_PROBE_MATCH:
            doc, header, body = self.createProbeMatchMessage(env)
        elif action == ACTION_RESOLVE:
            doc, header, body = self.createResolveMessage(env)
        elif action == ACTION_RESOLVE_MATCH:
            doc, header, body = self.createResolveMatchMessage(env)
        elif action == ACTION_HELLO:
            doc, header, body = self.createHelloMessage(env)
        elif action == ACTION_BYE:
            doc, header, body = self.createByeMessage(env)
        else:
            raise ValueError(f'unknown action {action}')
        return tostring(doc)

    def createProbeMessage(self, env):
        doc, header, body = self._createSkelSoapMessage(env.getAction(), env.getMessageId(),
                                                       to=env.getTo(),
                                                       replyTo=env.getReplyTo())
        probeEl = SubElement(body, WsdTag('Probe'))
        self._createTypeNodes(probeEl, env.getTypes())
        self._createScopeNodes(probeEl, env.getScopes())
        return doc, header, body

    def createProbeMatchMessage(self, env):
        doc, header, body = self._createSkelSoapMessage(env.getAction(), env.getMessageId(),
                                                       relatesTo=env.getRelatesTo(),
                                                       to=env.getTo(),
                                                       replyTo=env.getReplyTo(),
                                                       appSequence=self._mkAppSequenceDict(env))
        probeMatchesEl = SubElement(body, WsdTag('ProbeMatches'))

        probeMatches = env.getProbeResolveMatches()
        for probeMatch in probeMatches:
            probeMatchEl = SubElement(probeMatchesEl, WsdTag('ProbeMatch'))
            self._createEprNode(probeMatchEl, probeMatch.getEPR())
            self._createTypeNodes(probeMatchEl, probeMatch.getTypes())
            self._createScopeNodes(probeMatchEl, probeMatch.getScopes())
            self._createXAddrNodes(probeMatchEl, probeMatch.getXAddrs())
            self.mkSubElementWithText(probeMatchEl, WsdTag('MetadataVersion'), probeMatch.getMetadataVersion())
        return doc, header, body

    def createResolveMessage(self, env):
        doc, header, body = self._createSkelSoapMessage(env.getAction(), env.getMessageId(),
                                                        to=env.getTo(),
                                                        replyTo=env.getReplyTo())
        resolveEl = SubElement(body, WsdTag('Resolve'))
        self._createEprNode(resolveEl, env.getEPR())
        return doc, header, body

    def createResolveMatchMessage(self, env):
        doc, header, body = self._createSkelSoapMessage(env.getAction(),
                                                       env.getMessageId(),
                                                       relatesTo=env.getRelatesTo(),
                                                       to=env.getTo(),
                                                       appSequence=self._mkAppSequenceDict(env))
        resolveMatchesEl = SubElement(body, WsdTag('ResolveMatches'))
        if len(env.getProbeResolveMatches()) > 0:
            resolveMatch = env.getProbeResolveMatches()[0]
            resolveMatchEl = SubElement(resolveMatchesEl, WsdTag('ResolveMatch'))
            self._createEprNode(resolveMatchEl, resolveMatch.getEPR())
            self._createTypeNodes(resolveMatchEl, resolveMatch.getTypes())
            self._createScopeNodes(resolveMatchEl, resolveMatch.getScopes())
            self._createXAddrNodes(resolveMatchEl, resolveMatch.getXAddrs())
            self.mkSubElementWithText(resolveMatchEl, WsdTag('MetadataVersion'), resolveMatch.getMetadataVersion())
        return doc, header, body

    def createHelloMessage(self, env):
        doc, header, body = self._createSkelSoapMessage(env.getAction(), env.getMessageId(),
                                                       to=env.getTo(),
                                                       relatesTo=env.getRelatesTo(),
                                                       relatesToAttrib={"RelationshipType": "d:Suppression"},
                                                       appSequence=self._mkAppSequenceDict(env))
        helloEl = SubElement(body, WsdTag('Hello'))
        self._createEprNode(helloEl, env.getEPR())
        self._createTypeNodes(helloEl, env.getTypes())
        self._createScopeNodes(helloEl, env.getScopes())
        self._createXAddrNodes(helloEl, env.getXAddrs())
        self.mkSubElementWithText(helloEl, WsdTag('MetadataVersion'), env.getMetadataVersion())
        return doc, header, body

    def createByeMessage(self, env):
        doc, header, body = self._createSkelSoapMessage(env.getAction(), env.getMessageId(),
                                                       to=env.getTo(),
                                                       appSequence=self._mkAppSequenceDict(env))
        byeEl = SubElement(body, WsdTag('Bye'))
        self._createEprNode(byeEl, env.getEPR())
        return doc, header, body

    def _createSkelSoapMessage(self, soapAction, messageId,
                               relatesTo=None,
                               relatesToAttrib=None,
                               to=None,
                               replyTo=None,
                               appSequence=None):
        doc = Element(S12Tag('Envelope'),
                      nsmap=_namespaces_map)  # Prefix.partialMap(Prefix.S12, Prefix.WSA, Prefix.WSD, Prefix.DPWS))
        header = SubElement(doc, S12Tag('Header'))
        action = SubElement(header, WsaTag('Action'))
        action.text = soapAction
        body = SubElement(doc, S12Tag('Body'))

        self.mkSubElementWithText(header, WsaTag('MessageID'), messageId)
        if relatesTo:
            self.mkSubElementWithText(header, WsaTag('RelatesTo'), text=relatesTo, attrib=relatesToAttrib)
        if to:
            self.mkSubElementWithText(header, WsaTag('To'), to)
        if replyTo:
            self.mkSubElementWithText(header, WsaTag('ReplyTo'), replyTo)
        if appSequence:
            SubElement(header, WsdTag('AppSequence'), attrib=appSequence)
        return doc, header, body

    @staticmethod
    def _mkAppSequenceDict(env):
        return {"InstanceId": env.getInstanceId(),
                "MessageNumber": env.getMessageNumber()}

    def _createTypeNodes(self, parentNode, types):
        if types is not None and len(types) > 0:
            ns_map = {}
            typeList = []

            for i, _type in enumerate(types):
                ns, localname = _type.namespace, _type.localname
                prefix = _getPrefix(parentNode, ns)
                if prefix is None:
                    prefix = getRandomStr()
                    ns_map[prefix] = ns
                if i == 0:
                    # make namespace of first type the default namespace (so that we can test handling of this case)
                    ns_map[None] = ns
                    typeList.append(localname)
                else:
                    typeList.append(prefix + ":" + localname)

            typesString = " ".join(typeList)
            self.mkSubElementWithText(parentNode, WsdTag('Types'), typesString, nsmap=ns_map)

    def _createScopeNodes(self, parentNode, scopes):
        if scopes is not None and len(scopes) > 0:
            scopesString = " ".join([x.getQuotedValue() for x in scopes])
            self.mkSubElementWithText(parentNode, WsdTag('Scopes'), scopesString)

    def _createXAddrNodes(self, parentNode, xAddrs):
        if xAddrs is not len(xAddrs) > 0:
            addrString = " ".join([x for x in xAddrs])
            self.mkSubElementWithText(parentNode, WsdTag('XAddrs'), addrString)

    def _createEprNode(self, parentNode, epr):
        eprEl = SubElement(parentNode, WsaTag("EndpointReference"))
        self.mkSubElementWithText(eprEl, WsaTag('Address'), epr)

    @staticmethod
    def mkSubElementWithText(parentNode, qname, text, attrib=None, nsmap=None):
        elem = SubElement(parentNode, qname, attrib=attrib or {}, nsmap=nsmap or {})
        elem.text = text
        return elem


def extractSoapUdpAddressFromURI(uri):
    val = uri.getPathExQueryFragment().split(":")
    part1 = val[0][2:]
    part2 = None
    if val[1].count('/') > 0:
        part2 = int(val[1][:val[1].index('/')])
    else:
        part2 = int(val[1])
    addr = [part1, part2]
    return addr


class AddressMonitorThread(threading.Thread):
    """ This thread frequently checks the available Network adapters.
    Any change is reported vis wsd._networkAddressRemoved or wsd._networkAddressAdded
    """

    def __init__(self, wsd):
        self._addrs = set()
        self._wsd = wsd
        self._logger = logging.getLogger('sdc.discover.monitor')
        self._quitEvent = threading.Event()
        super().__init__(name='AddressMonitorThread')
        self.daemon = True
        self._updateAddrs()

    def _updateAddrs(self):
        addrs = set(get_ipv4_addresses())

        disappeared = self._addrs.difference(addrs)
        new = addrs.difference(self._addrs)

        for addr in disappeared:
            self._wsd._networkAddressRemoved(addr)

        for addr in new:
            try:
                self._wsd._networkAddressAdded(addr)
            except:
                self._logger.warning(traceback.format_exc())
        self._addrs = addrs

    def run(self):
        try:
            while not self._quitEvent.wait(_NETWORK_ADDRESSES_CHECK_TIMEOUT):
                self._updateAddrs()
        except Exception:
            self._logger.error('Unhandled Exception at thread runtime. Thread will abort! %s',
                               traceback.format_exc())
            raise

    def schedule_stop(self):
        """Schedule stopping the thread.
        Use join() to wait, until thread really has been stopped
        """
        self._quitEvent.set()


@dataclass(frozen=True)
class _SocketPair:
    multi_in: socket.socket
    multi_out_uni_in: socket.socket


class NetworkingThreadWindows:
    """ Has one thread for sending and one for receiving"""

    @dataclass(order=True)
    class _EnqueuedMessage:
        send_time: float
        msg: Any = field(compare=False)

    def __init__(self, observer, logger, multicast_port, message_creator):
        self._observer = observer
        self._logger = logger
        self.multicast_port = multicast_port
        self.message_creator = message_creator
        self._recvThread = None
        self._qread_thread = None
        self._sendThread = None
        self._quitRecvEvent = threading.Event()
        self._quitSendEvent = threading.Event()
        self._send_queue = queue.PriorityQueue(10000)
        self._read_queue = queue.Queue(10000)
        self._knownMessageIds = deque(maxlen=50)

        self._select_in = []
        self._full_selector = selectors.DefaultSelector()
        self._sockets_by_address = {}
        self._sockets_by_address_lock = threading.RLock()
        self._uni_out_socket = None

    def _register(self, sock):
        self._select_in.append(sock)
        self._full_selector.register(sock, selectors.EVENT_READ)

    def _unregister(self, sock):
        self._select_in.remove(sock)
        self._full_selector.unregister(sock)

    @staticmethod
    def _makeMreq(addr):
        return struct.pack("4s4s", socket.inet_aton(MULTICAST_IPV4_ADDRESS), socket.inet_aton(addr))

    @staticmethod
    def _createMulticastOutSocket(addr):
        sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        sock.setsockopt(socket.IPPROTO_IP, socket.IP_MULTICAST_TTL, MULTICAST_OUT_TTL)
        if addr is None:
            sock.setsockopt(socket.IPPROTO_IP, socket.IP_MULTICAST_IF, socket.INADDR_ANY)
        else:
            _addr = socket.inet_aton(addr)
            sock.setsockopt(socket.IPPROTO_IP, socket.IP_MULTICAST_IF, _addr)
        return sock

    @staticmethod
    def _createMulticastInSocket(addr, port):
        sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM, socket.IPPROTO_UDP)
        sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        sock.bind((addr, port))
        sock.setblocking(False)
        return sock

    def addSourceAddr(self, addr):
        """None means 'system default'"""
        multicast_in_sock = self._createMulticastInSocket(addr, self.multicast_port)
        try:
            multicast_in_sock.setsockopt(socket.IPPROTO_IP, socket.IP_ADD_MEMBERSHIP, self._makeMreq(addr))
        except socket.error:  # if 1 interface has more than 1 address, exception is raised for the second
            print(traceback.format_exc())
            pass
        multicast_out_sock = self._createMulticastOutSocket(addr)
        with self._sockets_by_address_lock:
            self._register(multicast_out_sock)
            self._register(multicast_in_sock)
            self._sockets_by_address[addr] = _SocketPair(multicast_in_sock, multicast_out_sock)

    def removeSourceAddr(self, addr):
        sock_pair = self._sockets_by_address.get(addr)
        if sock_pair:
            with self._sockets_by_address_lock:
                for sock in (sock_pair.multi_in, sock_pair.multi_out_uni_in):
                    self._unregister(sock)
                    sock.close()
                del self._sockets_by_address[addr]

    def addUnicastMessage(self, env, addr, port, initialDelay=0):
        msg = Message(env, addr, port, Message.UNICAST, initialDelay)
        self._logger.debug(
            'addUnicastMessage: adding message Id %s. delay=%.2f', env.getMessageId(), initialDelay)
        self._repeated_enqueue_msg(msg, initialDelay, UNICAST_UDP_REPEAT, UNICAST_UDP_MIN_DELAY,
                                   UNICAST_UDP_MAX_DELAY, UNICAST_UDP_UPPER_DELAY)

    def addMulticastMessage(self, env, addr, port, initialDelay=0):
        msg = Message(env, addr, port, Message.MULTICAST, initialDelay)
        self._logger.debug(
            'addMulticastMessage: adding message Id %s. delay=%.2f', env.getMessageId(), initialDelay)
        self._repeated_enqueue_msg(msg, initialDelay, MULTICAST_UDP_REPEAT, MULTICAST_UDP_MIN_DELAY,
                                   MULTICAST_UDP_MAX_DELAY, MULTICAST_UDP_UPPER_DELAY)

    def _repeated_enqueue_msg(self, msg, initial_delay_ms, repeat, min_delay_ms, max_delay_ms, upper_delay_ms):
        if not self._quitSendEvent.is_set():
            next_send = time.time() + initial_delay_ms / 1000.0
            dt = random.randrange(min_delay_ms, max_delay_ms) / 1000.0  # millisec -> seconds
            self._send_queue.put(self._EnqueuedMessage(next_send, msg))
            for _ in range(repeat):
                next_send += dt
                self._send_queue.put(self._EnqueuedMessage(next_send, msg))
                dt = min(dt * 2, upper_delay_ms)

    def _run_send(self):
        """send-loop"""
        while not self._quitSendEvent.is_set() or not self._send_queue.empty():
            if self._send_queue.empty():
                time.sleep(SEND_LOOP_IDLE_SLEEP)  # nothing to do currently
            else:
                if self._send_queue.queue[0].send_time <= time.time():
                    enqueued_msg = self._send_queue.get()
                    self._sendMsg(enqueued_msg.msg)
                else:
                    time.sleep(SEND_LOOP_BUSY_SLEEP)  # this creates a 10ms raster for sending, but that is good enough

    def _run_recv(self):
        """ run by thread"""
        while not self._quitRecvEvent.is_set():
            if len(self._sockets_by_address) == 0:
                # avoid errors while no sockets are registered
                time.sleep(0.1)
                continue
            try:
                self._recv_messages()
            except:
                if not self._quitRecvEvent.is_set():  # only log error if it does not happen during stop
                    self._logger.error('_run_recv:%s', traceback.format_exc())

    def isFromMySocket(self, addr):
        with self._sockets_by_address_lock:
            for ip_addr, sock_pair in self._sockets_by_address.items():
                if addr[0] == ip_addr:
                    try:
                        sock_name = sock_pair.multi_out_uni_in.getsockname()
                        if addr[1] == sock_name[1]:  # compare ports
                            return True
                    except OSError:  # port is not opened
                        continue
        return False

    def _recv_messages(self):
        """For performance reasons this thread only writes to a queue, no parsing etc."""
        for key, events in self._full_selector.select(timeout=0.1):
            sock = key.fileobj
            try:
                data, addr = sock.recvfrom(BUFFER_SIZE)
            except socket.error as e:
                self._logger.warning('socket read error %s', e)
                time.sleep(0.01)
                continue
            if self.isFromMySocket(addr):
                continue
            self._add_to_recv_queue(addr, data)

    def _add_to_recv_queue(self, addr, data):
        # method is needed for testing
        self._read_queue.put((addr, data))

    def _run_q_read(self):
        """Read from internal queue and process message"""
        while not self._quitRecvEvent.is_set():
            try:
                try:
                    incoming = self._read_queue.get(timeout=0.1)
                except queue.Empty:
                    pass
                else:
                    addr, data = incoming
                    getCommunicationLogger().logDiscoveryMsgIn(addr[0], data)

                    env = parseEnvelope(data, addr[0], self._logger)
                    if env is None:  # fault or failed to parse
                        continue

                    mid = env.getMessageId()
                    if mid in self._knownMessageIds:
                        self._logger.debug('message Id %s already known. This is a duplicate receive, ignoring.', mid)
                        continue
                    else:
                        self._knownMessageIds.appendleft(mid)
                    self._observer.envReceived(env, addr)
            except:
                self._logger.error('unexpected error in queue read thread: %s', traceback.format_exc())
        self._logger.info('queue read thread terminates')

    def _sendMsg(self, msg):
        action = msg._env.getAction().split('/')[-1]  # only last part
        if action in ('ResolveMatches', 'ProbeMatches'):
            self._logger.debug('_sendMsg: sending %s %s to %s ProbeResolveMatches=%r, epr=%s, msgNo=%r',
                               action,
                               msg.msgType(),
                               msg.getAddr(),
                               msg._env.getProbeResolveMatches(),
                               msg._env.getEPR(),
                               msg._env._messageNumber
                               )
        elif action == 'Probe':
            self._logger.debug('_sendMsg: sending %s %s to %s types=%s scopes=%r',
                               action,
                               msg.msgType(),
                               msg.getAddr(),
                               types_info(msg._env.getTypes()),
                               msg._env.getScopes(),
                               )
        else:
            self._logger.debug('_sendMsg: sending %s %s to %s xaddr=%r, epr=%s, msgNo=%r',
                               action,
                               msg.msgType(),
                               msg.getAddr(),
                               msg._env.getXAddrs(),
                               msg._env.getEPR(),
                               msg._env._messageNumber
                               )

        data = self.message_creator.createMessage(msg.getEnv())

        if msg.msgType() == Message.UNICAST:
            getCommunicationLogger().logDiscoveryMsgOut(msg.getAddr(), data)
            self._uniOutSocket.sendto(data, (msg.getAddr(), msg.getPort()))
        else:
            getCommunicationLogger().logBroadCastMsgOut(data)
            with self._sockets_by_address_lock:
                for sock_pair in self._sockets_by_address.values():
                    sock_pair.multi_out_uni_in.sendto(data, (msg.getAddr(), msg.getPort()))

    def start(self):
        self._logger.debug('%s: starting ', self.__class__.__name__)
        self._uniOutSocket = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)

        self._recvThread = threading.Thread(target=self._run_recv, name='wsd.recvThread')
        self._qread_thread = threading.Thread(target=self._run_q_read, name='wsd.qreadThread')
        self._sendThread = threading.Thread(target=self._run_send, name='wsd.sendThread')
        self._recvThread.daemon = True
        self._qread_thread.daemon = True
        self._sendThread.daemon = True
        self._recvThread.start()
        self._qread_thread.start()
        self._sendThread.start()

    def schedule_stop(self):
        """Schedule stopping the thread.
        Use join() to wait, until thread really has been stopped
        """
        self._logger.debug('%s: schedule_stop ', self.__class__.__name__)
        self._quitRecvEvent.set()
        self._quitSendEvent.set()

    def join(self):
        self._logger.debug('%s: join... ', self.__class__.__name__)
        self._recvThread.join(1)
        self._qread_thread.join(1)
        self._sendThread.join(10)
        self._recvThread = None
        self._qread_thread = None
        self._sendThread = None
        for sock in self._select_in:
            sock.close()
        self._uniOutSocket.close()
        self._full_selector.close()
        self._logger.debug('%s: ... join done', self.__class__.__name__)

    def getActiveAddresses(self):
        with self._sockets_by_address_lock:
            return list(self._sockets_by_address.keys())


@dataclass(frozen=True)
class _Sockets:
    multi_in: socket.socket
    uni_in: socket.socket
    multi_out_uni_in: socket.socket


class NetworkingThreadPosix(NetworkingThreadWindows):

    @staticmethod
    def _createMulticastInSocket(addr, port):
        sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM, socket.IPPROTO_UDP)
        sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        sock.bind((MULTICAST_IPV4_ADDRESS, port))
        sock.setblocking(False)
        mreq = struct.pack("4s4s", socket.inet_aton(MULTICAST_IPV4_ADDRESS), socket.inet_aton(addr))
        sock.setsockopt(socket.IPPROTO_IP, socket.IP_ADD_MEMBERSHIP, mreq)
        return sock

    @staticmethod
    def _createUnicastInSocket(addr, port):
        sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM, socket.IPPROTO_UDP)
        sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        sock.bind((addr, port))
        sock.setblocking(False)
        return sock

    def addSourceAddr(self, addr):
        """None means 'system default'"""
        try:
            multicast_in_sock = self._createMulticastInSocket(addr, self.multicast_port)
        except socket.error:  # if 1 interface has more than 1 address, exception is raised for the second
            print(traceback.format_exc())
            pass
        unicast_in_sock = self._createUnicastInSocket(addr,
                                                      self.multicast_port)  # allows handling of unicast messages on self.multicast_port
        multicast_out_sock = self._createMulticastOutSocket(addr)

        with self._sockets_by_address_lock:
            self._register(multicast_out_sock)
            self._register(unicast_in_sock)
            self._register(multicast_in_sock)
            self._sockets_by_address[addr] = _Sockets(multicast_in_sock, unicast_in_sock, multicast_out_sock)

    def removeSourceAddr(self, addr):
        sockets = self._sockets_by_address.get(addr)
        if sockets:
            with self._sockets_by_address_lock:
                for sock in (sockets.multi_in, sockets.uni_in, sockets.multi_out_uni_in):
                    self._unregister(sock)
                    sock.close()
                del self._sockets_by_address[addr]


class Message:
    MULTICAST = 'multicast'
    UNICAST = 'unicast'

    def __init__(self, env, addr, port, msgType, initialDelay=0):
        """msgType shall be Message.MULTICAST or Message.UNICAST"""
        self._env = env
        self._addr = addr
        self._port = port
        self._msgType = msgType

    def getEnv(self):
        return self._env

    def getAddr(self):
        return self._addr

    def getPort(self):
        return self._port

    def msgType(self):
        return self._msgType


class Service:
    def __init__(self, types, scopes, xAddrs, epr, instanceId, metadata_version=1):
        self._types = types
        self._scopes = scopes
        self._xAddrs = xAddrs
        self._epr = epr
        self._instanceId = instanceId
        self._messageNumber = 0
        self._metadataVersion = metadata_version

    def getTypes(self):
        return self._types

    def setTypes(self, types):
        self._types = types

    def getScopes(self):
        return self._scopes

    def setScopes(self, scopes):
        self._scopes = scopes

    def getXAddrs(self):
        ret = []
        ipAddrs = None
        for xAddr in self._xAddrs:
            if '{ip}' in xAddr:
                if ipAddrs is None:
                    ipAddrs = get_ipv4_addresses()
                for ipAddr in ipAddrs:
                    if ipAddr not in _IP_BLACKLIST:
                        ret.append(xAddr.format(ip=ipAddr))
            else:
                ret.append(xAddr)
        return ret

    def setXAddrs(self, xAddrs):
        self._xAddrs = xAddrs

    def getEPR(self):
        return self._epr

    def setEPR(self, epr):
        self._epr = epr

    def getInstanceId(self):
        return self._instanceId

    def setInstanceId(self, instanceId):
        self._instanceId = instanceId

    def getMessageNumber(self):
        return self._messageNumber

    def setMessageNumber(self, messageNumber):
        self._messageNumber = messageNumber

    def getMetadataVersion(self):
        return self._metadataVersion

    def setMetadataVersion(self, metadataVersion):
        self._metadataVersion = metadataVersion

    def incrementMetadataVersion(self):
        self._metadataVersion = self._metadataVersion + 1

    def incrementMessageNumber(self):
        self._messageNumber = self._messageNumber + 1

    def isLocatedOn(self, *ipaddresses):
        """
        :param ipaddresses: ip addresses, lists of strings or strings
        """
        my_addresses = []
        for i in ipaddresses:
            if isinstance(i, str):
                my_addresses.append(i)
            else:
                my_addresses.extend(i)
        for addr in self.getXAddrs():
            parsed = urllib.parse.urlsplit(addr)
            ip_addr = parsed.netloc.split(':')[0]
            if ip_addr in my_addresses:
                return True
        return False

    def __repr__(self):
        return 'Service epr={}, instanceId={} Xaddr={} scopes={} types={}'.format(self._epr, self._instanceId,
                                                                                  self._xAddrs,
                                                                                  ', '.join(
                                                                                      [str(x) for x in self._scopes]),
                                                                                  ', '.join(
                                                                                      [str(x) for x in self._types]))

    def __str__(self):
        return 'Service epr={}, instanceId={}\n   Xaddr={}\n   scopes={}\n   types={}'.format(self._epr,
                                                                                              self._instanceId,
                                                                                              self._xAddrs,
                                                                                              ', '.join([str(x) for x in
                                                                                                         self._scopes]),
                                                                                              ', '.join([str(x) for x in
                                                                                                         self._types]))


def _isTypeInList(ttype, types):
    for entry in types:
        if matchType(ttype, entry):
            return True
    return False


def _isScopeInList(scope, scopes):
    for entry in scopes:
        if matchScope(scope.getValue(), entry.getValue(), scope.getMatchBy()):
            return True
    return False


def _matchesFilter(service, types, scopes, logger=None):
    if types is not None:
        srv_ty = service.getTypes()
        for ttype in types:
            if not _isTypeInList(ttype, srv_ty):
                if logger:
                    logger.debug('types not matching: {} is not in types list {}'.format(ttype, srv_ty))
                return False
        if logger:
            logger.debug('matching types')
    if scopes is not None:
        srv_sc = service.getScopes()
        for scope in scopes:
            if not _isScopeInList(scope, srv_sc):
                if logger:
                    logger.debug('scope not matching: {} is not in scopes list {}'.format(scope, srv_sc))
                return False
        if logger:
            logger.debug('matching scopes')
    return True


def filterServices(services, types, scopes, logger=None):
    return [service for service in services if _matchesFilter(service, types, scopes, logger)]


class WSDiscoveryBase():
    """
    UDP based discovery.
    """
    # these flags control which data is included in ProbeResponse messages.
    PROBEMATCH_EPR = True
    PROBEMATCH_TYPES = True
    PROBEMATCH_SCOPES = True
    PROBEMATCH_XADDRS = True

    SoapEnvelopeCls = SoapEnvelope
    ServiceCls = Service
    MessageCreatorCls = MessageCreator

    def __init__(self, logger=None, multicast_port=None):
        """
        :param logger: use this logger. if None a logger 'sdc.discover' is created.
        :param multicast_port: a port number; if None, the default MULTICAST_PORT is used
        """
        self._logger = logger or logging.getLogger('sdc.discover')
        self.multicast_port = multicast_port or MULTICAST_PORT
        self._networkingThread = None
        self._addrsMonitorThread = None
        self._serverStarted = False
        self._remoteServices = {}
        self._localServices = {}

        self._dpActive = False  # True if discovery proxy detected (is not relevant in sdc context)
        self._dpAddr = None
        self._dpEPR = None

        self._remoteServiceProbeMatchCallback = None
        self._remoteServiceHelloCallback = None
        self._remoteServiceHelloCallbackTypesFilter = None
        self._remoteServiceHelloCallbackScopesFilter = None
        self._remoteServiceByeCallback = None
        self._remoteServiceResolveMatchCallback = None  # B.D.
        self._onProbeCallback = None
        random.seed(int(time.time() * 1000000))

    def setRemoteServiceProbeMatchCallback(self, cb):
        """Set callback, which will be called when a service was received via a ProbeMatch message.
        Service is passed as a parameter to the callback
        Set None to disable callback
        """
        self._remoteServiceProbeMatchCallback = cb

    def setRemoteServiceHelloCallback(self, cb, types=None, scopes=None):
        """Set callback, which will be called when new service appeared online
        and sent Hi message

        typesFilter and scopesFilter might be list of types and scopes.
        If filter is set, callback is called only for Hello messages,
        which match filter

        Set None to disable callback
        """
        self._remoteServiceHelloCallback = cb
        self._remoteServiceHelloCallbackTypesFilter = types
        self._remoteServiceHelloCallbackScopesFilter = scopes

    def setRemoteServiceByeCallback(self, cb):
        """Set callback, which will be called when a service goes offline and sent a Bye message
        Set None to disable callback
        """
        self._remoteServiceByeCallback = cb

    def setRemoteServiceResolveMatchCallback(self, cb):  # B.D.
        self._remoteServiceResolveMatchCallback = cb

    def setOnProbeCallback(self, cb):
        self._onProbeCallback = cb

    def _addRemoteService(self, service):
        epr = service.getEPR()
        if not epr:
            self._logger.info('service without epr, ignoring it! %r', service)
            return
        s = self._remoteServices.get(service.getEPR())
        if not s:
            self._remoteServices[service.getEPR()] = service
            self._logger.info('new remote %r', service)
        else:
            if service.getMetadataVersion() == s.getMetadataVersion():
                self._logger.debug('_addRemoteService: remote Service %s:\n    MetadataVersion: %d',
                                   service.getEPR(), service.getMetadataVersion())
                merged = []
                if len(service.getXAddrs()) > len(s.getXAddrs()):
                    s.setXAddrs(service.getXAddrs())
                    merged.append('XAddr={}'.format(service.getXAddrs()))

                if len(service.getScopes()) > len(s.getScopes()):
                    s.setScopes(service.getScopes())
                    merged.append('Scopes={}'.format(service.getScopes()))

                if len(service.getTypes()) > len(s.getTypes()):
                    s.setTypes(service.getTypes())
                    merged.append('Types={}'.format(service.getTypes()))
                if merged:
                    self._logger.info('merge from remote Service %s:\n      %r',
                                      service.getEPR(), '\n      '.join(merged))
            elif service.getMetadataVersion() > s.getMetadataVersion():
                self._logger.info('remote Service %s:\n    updated MetadataVersion\n      '
                                  'updated: %d\n      existing: %d',
                                  service.getEPR(), service.getMetadataVersion(), s.getMetadataVersion())
                self._remoteServices[service.getEPR()] = service
            else:
                self._logger.debug('_addRemoteService: remote Service %s:\n    outdated MetadataVersion\n      '
                                   'outdated: %d\n      existing: %d',
                                   service.getEPR(), service.getMetadataVersion(), s.getMetadataVersion())

    def _removeRemoteService(self, epr):
        if epr in self._remoteServices:
            del self._remoteServices[epr]

    def handleEnv(self, env, addr):
        act = env.getAction()
        self._logger.debug('handleEnv: received %s from %s', act.split('/')[-1], addr)
        if act == ACTION_PROBE_MATCH:
            for match in env.getProbeResolveMatches():
                service = self.ServiceCls(match.getTypes(), match.getScopes(), match.getXAddrs(), match.getEPR(),
                                          env.getInstanceId(), metadata_version=int(match.getMetadataVersion()))
                self._addRemoteService(service)
                if match.getXAddrs() is None or len(match.getXAddrs()) == 0:
                    self._logger.info('%s(%s) has no Xaddr, sending resolve message', match.getEPR(), addr)
                    self._sendResolve(match.getEPR())
                elif not match.getTypes():
                    self._logger.info('%s(%s) has no Types, sending resolve message', match.getEPR(), addr)
                    self._sendResolve(match.getEPR())
                elif not match.getScopes():
                    self._logger.info('%s(%s) has no Scopes, sending resolve message', match.getEPR(), addr)
                    self._sendResolve(match.getEPR())

                if self._remoteServiceProbeMatchCallback is not None:
                    self._remoteServiceProbeMatchCallback(addr, service)

        elif act == ACTION_RESOLVE_MATCH:
            for match in env.getProbeResolveMatches():
                service = self.ServiceCls(match.getTypes(), match.getScopes(), match.getXAddrs(), match.getEPR(),
                                          env.getInstanceId(), metadata_version=int(match.getMetadataVersion()))
                self._addRemoteService(service)
                if self._remoteServiceResolveMatchCallback is not None:
                    self._remoteServiceResolveMatchCallback(service)

        elif act == ACTION_PROBE:
            services = filterServices(self._localServices.values(), env.getTypes(), env.getScopes())
            if services:
                self._sendProbeMatch(services, env.getMessageId(), addr)
            if self._onProbeCallback is not None:
                self._onProbeCallback(addr, env)

        elif act == ACTION_RESOLVE:
            if env.getEPR() in self._localServices:
                service = self._localServices[env.getEPR()]
                self._sendResolveMatch(service, env.getMessageId(), addr)

        elif act == ACTION_HELLO:
            # check if it is from a discovery proxy
            rt = env.getRelationshipType()
            if rt is not None and rt.localname == "Suppression" and rt.namespace == NS_D:
                xAddr = env.getXAddrs()[0]
                # only support 'soap.udp'
                if xAddr.startswith("soap.udp:"):
                    self._dpActive = True
                    self._dpAddr = extractSoapUdpAddressFromURI(URI(xAddr))
                    self._dpEPR = env.getEPR()

            service = self.ServiceCls(env.getTypes(), env.getScopes(), env.getXAddrs(), env.getEPR(),
                                      env.getInstanceId(),
                                      metadata_version=int(env.getMetadataVersion()))
            self._addRemoteService(service)
            if not env.getXAddrs():  # B.D.
                self._logger.debug('%s(%s) has no Xaddr, sending resolve message', env.getEPR(), addr)
                self._sendResolve(env.getEPR())
            if self._remoteServiceHelloCallback is not None:
                if _matchesFilter(service,
                                  self._remoteServiceHelloCallbackTypesFilter,
                                  self._remoteServiceHelloCallbackScopesFilter):
                    self._remoteServiceHelloCallback(addr, service)

        elif act == ACTION_BYE:
            # if the bye is from discovery proxy... revert back to multicasting
            if self._dpActive and self._dpEPR == env.getEPR():
                self._dpActive = False
                self._dpAddr = None
                self._dpEPR = None

            self._removeRemoteService(env.getEPR())
            if self._remoteServiceByeCallback is not None:
                self._remoteServiceByeCallback(addr, env.getEPR())
        else:
            self._logger.info('unknown action %s', act)

    def envReceived(self, env, addr):
        self.handleEnv(env, addr)

    def _mkResolveMatchEnvelope(self, service, relatesTo):
        env = self.SoapEnvelopeCls()
        env.setAction(ACTION_RESOLVE_MATCH)
        env.setTo(WSA_ANONYMOUS)
        env.setInstanceId(str(service.getInstanceId()))
        env.setMessageNumber(str(service.getMessageNumber()))
        env.setRelatesTo(relatesTo)
        env.setProbeResolveMatches([ProbeResolveMatch(service.getEPR(), service.getTypes(), service.getScopes(),
                                                      service.getXAddrs(), str(service.getMetadataVersion()))])
        return env

    def _sendResolveMatch(self, service, relatesTo, addr):
        self._logger.info('sending resolve match to %s', addr)
        service.incrementMessageNumber()
        env = self._mkResolveMatchEnvelope(service, relatesTo)
        self._networkingThread.addUnicastMessage(env, addr[0], addr[1])

    def _mkProbeMatchEnvelope(self, service, relatesTo, msgNumber):
        env = self.SoapEnvelopeCls()
        env.setAction(ACTION_PROBE_MATCH)
        env.setTo(WSA_ANONYMOUS)
        env.setInstanceId(self.generate_instance_id())
        env.setMessageNumber(str(msgNumber))
        env.setRelatesTo(relatesTo)

        # add values to ProbeResponse acc. to flags
        epr = service.getEPR() if self.PROBEMATCH_EPR else ''
        types = service.getTypes() if self.PROBEMATCH_TYPES else []
        scopes = service.getScopes() if self.PROBEMATCH_SCOPES else []
        xaddrs = service.getXAddrs() if self.PROBEMATCH_XADDRS else []
        env.setProbeResolveMatches([ProbeResolveMatch(epr, types, scopes, xaddrs,
                                                      str(service.getMetadataVersion()))])
        return env

    def _sendProbeMatch(self, services, relatesTo, addr):
        self._logger.info('sending probe match to %s for %d services', addr, len(services))
        msgNumber = 1
        # send one match response for every service, dpws explorer can't handle telegram otherwise if too many devices reported
        for service in services:
            self._logger.debug('sending probe match 1')
            env = self._mkProbeMatchEnvelope(service, relatesTo, msgNumber)
            self._logger.debug('sending probe match 2')
            self._networkingThread.addUnicastMessage(env, addr[0], addr[1], random.randint(0, APP_MAX_DELAY))

    def _mkProbeEnvelope(self, types, scopes):
        env = self.SoapEnvelopeCls()
        env.setAction(ACTION_PROBE)
        env.setTo(ADDRESS_ALL)
        env.setTypes(types)
        env.setScopes(scopes)
        return env

    def _sendProbe(self, types=None, scopes=None):
        self._logger.debug('sending probe types=%r scopes=%r', types_info(types), scopes)
        env = self._mkProbeEnvelope(types, scopes)
        if self._dpActive:
            self._networkingThread.addUnicastMessage(env, self._dpAddr[0], self._dpAddr[1])
        else:
            self._networkingThread.addMulticastMessage(env, MULTICAST_IPV4_ADDRESS, self.multicast_port)

    def _mkResolveEnvelope(self, epr):
        env = self.SoapEnvelopeCls()
        env.setAction(ACTION_RESOLVE)
        env.setTo(ADDRESS_ALL)
        env.setEPR(epr)
        return env

    def _sendResolve(self, epr):
        self._logger.debug('sending resolve on %s', epr)
        env = self._mkResolveEnvelope(epr)
        if self._dpActive:
            self._networkingThread.addUnicastMessage(env, self._dpAddr[0], self._dpAddr[1])
        else:
            self._networkingThread.addMulticastMessage(env, MULTICAST_IPV4_ADDRESS, self.multicast_port)

    def _mkHelloEnvelope(self, service):
        env = self.SoapEnvelopeCls()
        env.setAction(ACTION_HELLO)
        env.setTo(ADDRESS_ALL)
        env.setInstanceId(str(service.getInstanceId()))
        env.setMessageNumber(str(service.getMessageNumber()))
        env.setTypes(service.getTypes())
        env.setScopes(service.getScopes())
        env.setXAddrs(service.getXAddrs())
        env.setEPR(service.getEPR())
        env.setMetadataVersion(str(service.getMetadataVersion()))
        return env

    def _sendHello(self, service):
        self._logger.info('sending hello on %s', service)
        service.incrementMessageNumber()
        env = self._mkHelloEnvelope(service)
        self._networkingThread.addMulticastMessage(env, MULTICAST_IPV4_ADDRESS, self.multicast_port,
                                                   random.randint(0, APP_MAX_DELAY))

    def _mkByeEnvelope(self, service):
        env = self.SoapEnvelopeCls()
        env.setAction(ACTION_BYE)
        env.setTo(ADDRESS_ALL)
        env.setInstanceId(str(service.getInstanceId()))
        env.setMessageNumber(str(service.getMessageNumber()))
        env.setEPR(service.getEPR())
        return env

    def _sendBye(self, service):
        self._logger.debug('sending bye on %s', service)
        env = self._mkByeEnvelope(service)
        service.incrementMessageNumber()
        self._networkingThread.addMulticastMessage(env, MULTICAST_IPV4_ADDRESS, self.multicast_port)

    def start(self):
        """start the discovery server - should be called before using other functions"""
        if not self._serverStarted:
            self._startThreads()
            self._serverStarted = True

    def stop(self):
        """cleans up and stops the discovery server"""
        if self._serverStarted:
            self.clearRemoteServices()
            self.clearLocalServices()

            self._stopThreads()
            self._serverStarted = False

    def _isAcceptedAddress(self, addr):  # pylint: disable=unused-argument
        """ accept any interface. Overwritten in derived classes."""
        return True

    def _networkAddressAdded(self, addr):
        if not self._isAcceptedAddress(addr):
            self._logger.debug('network Address ignored: %s', addr)
            return

        self._logger.debug('network Address Add: %s', addr)
        try:
            self._networkingThread.addSourceAddr(addr)
            for service in self._localServices.values():
                self._sendHello(service)
        except:
            self._logger.warning('error in network Address "%s" Added: %s', addr, traceback.format_exc())

    def _networkAddressRemoved(self, addr):
        self._logger.debug('network Address removed %s', addr)
        self._networkingThread.removeSourceAddr(addr)

    def _startThreads(self):
        if self._networkingThread is not None:
            return
        message_creator = self.MessageCreatorCls()
        if platform.system() != 'Windows':
            self._networkingThread = NetworkingThreadPosix(self, self._logger, self.multicast_port, message_creator)
        else:
            self._networkingThread = NetworkingThreadWindows(self, self._logger, self.multicast_port, message_creator)
        self._networkingThread.start()

        self._addrsMonitorThread = AddressMonitorThread(self)
        self._addrsMonitorThread.start()

    def _stopThreads(self):
        if self._networkingThread is None:
            return

        self._networkingThread.schedule_stop()
        self._addrsMonitorThread.schedule_stop()

        self._networkingThread.join()
        self._addrsMonitorThread.join()

        self._networkingThread = None
        self._addrsMonitorThread = None

    def clearRemoteServices(self):
        """clears remotely discovered services"""
        self._remoteServices.clear()

    def clearLocalServices(self):
        """send Bye messages for the services and remove them"""
        for service in self._localServices.values():
            self._sendBye(service)
        self._localServices.clear()

    def searchServices(self, types=None, scopes=None, timeout=5, repeatProbeInterval=3):
        """search for services given the TYPES and SCOPES in a given timeout
        :param repeatProbeInterval: send another probe message after x seconds"""
        if not self._serverStarted:
            raise Exception("Server not started")

        start = time.monotonic()
        end = start + timeout
        now = time.monotonic()
        while now < end:
            self._sendProbe(types, scopes)
            if now + repeatProbeInterval <= end:
                time.sleep(repeatProbeInterval)
            elif now < end:
                time.sleep(end - now)
            now = time.monotonic()
        return filterServices(self._remoteServices.values(), types, scopes)

    def searchMultipleTypes(self, typesList, scopes=None, timeout=10, repeatProbeInterval=3):
        """search for services given the list of TYPES and SCOPES in a given timeout.
        It returns services that match at least one of the types (OR condition).
        Can be used to search for devices that support Biceps Draft6 and Final with one search.
        :param repeatProbeInterval: send another probe message after x seconds"""
        if not self._serverStarted:
            raise Exception("Server not started")

        start = time.monotonic()
        end = start + timeout
        now = time.monotonic()
        while now < end:
            for t in typesList:
                self._sendProbe(t, scopes)
            now = time.monotonic()
            if now + repeatProbeInterval <= end:
                time.sleep(repeatProbeInterval)
            elif now < end:
                time.sleep(end - now)
        result = []
        for t in typesList:
            result.extend(filterServices(self._remoteServices.values(), t, scopes))
        return result

    def searchMedicalDeviceServicesinLocation(self, sdcLocation, timeout=3, bicepsVersion=None):
        if bicepsVersion is None:
            types = _FallbackMedicalDeviceTypesFilter
        else:
            types = bicepsVersion.MedicalDeviceTypesFilter
        services = self.searchServices(types=types, timeout=timeout)
        return sdcLocation.matchingServices(services)

    def publishService(self, epr, types, scopes, xAddrs):
        service = self._mkService(epr, types, scopes, xAddrs)
        self._sendHello(service)

    def _mkService(self, epr, types, scopes, xAddrs):
        """Publish a service with the given TYPES, SCOPES and XAddrs (service addresses)

        if xAddrs contains item, which includes {ip} pattern, one item per IP addres will be sent
        """
        if not self._serverStarted:
            raise Exception("Server not started")

        instanceId = self.generate_instance_id()
        metadata_version = self._localServices[epr].getMetadataVersion() + 1 if epr in self._localServices else 1
        service = self.ServiceCls(types, scopes, xAddrs, epr, instanceId, metadata_version=metadata_version)
        self._logger.info('publishing %r', service)
        self._localServices[epr] = service
        return service

    def clearService(self, epr):
        service = self._localServices[epr]
        self._sendBye(service)
        del self._localServices[epr]

    def getActiveAddresses(self):
        return self._networkingThread.getActiveAddresses()

    @staticmethod
    def generate_instance_id():
        return str(random.randint(1, 0xFFFFFFFF))


class WSDiscoveryBlacklist(WSDiscoveryBase):
    """ Binds to all IP addresses except the black listed ones. """

    def __init__(self, ignoredAdaptorIPAddresses=None, logger=None, multicast_port=None):
        """
        :param ignoredAdaptorIPAddresses: an optional list of (own) ip addresses that shall not be used for discovery.
                                          IP addresses are handled as regular expressions.
        """
        super(WSDiscoveryBlacklist, self).__init__(logger, multicast_port)
        tmp = [] if ignoredAdaptorIPAddresses is None else ignoredAdaptorIPAddresses
        self._ignoredAdaptorIPAddresses = [re.compile(x) for x in tmp]

    def _isAcceptedAddress(self, addr):
        """ check if any of the regular expressions matches the argument"""
        for x in self._ignoredAdaptorIPAddresses:
            if x.match(addr) is not None:
                return False
        return True


WSDiscovery = WSDiscoveryBlacklist  # deprecated name, for backward compatibility


class WSDiscoveryWhitelist(WSDiscoveryBase):
    """ Binds to all IP listed IP addresses. """

    def __init__(self, acceptedAdapterIPAddresses, logger=None, multicast_port=None):
        """
        :param acceptedAdaptorIPAddresses: an optional list of (own) ip addresses that shall not be used for discovery.
        """
        super(WSDiscoveryWhitelist, self).__init__(logger, multicast_port)
        tmp = [] if acceptedAdapterIPAddresses is None else acceptedAdapterIPAddresses
        self.acceptedAdapterIPAddresses = [re.compile(x) for x in tmp]

    def _isAcceptedAddress(self, addr):
        """ check if any of the regular expressions matches the argument"""
        for x in self.acceptedAdapterIPAddresses:
            if x.match(addr) is not None:
                return True
        return False


class WSDiscoverySingleAdapter(WSDiscoveryBase):
    """ Bind to a single adapter, identified by name.
    """

    def __init__(self, adapterName, logger=None, forceAdapterName=False, multicast_port=None):
        """
        :param adapterName: a string,  e.g. 'local area connection'.
                            parameter is only relevant if host has more than one adapter or forceName is True
                            If host has more than one adapter, the adapter with this friendly name is used, but if it does not exist, a RuntimeError is thrown.
        :param logger: use this logger. If none, 'sdc.discover' is used.
        :param forceAdapterName: if True, only this named adapter will be used.
                                 If False, and only one Adapter exists, the one existing adapter is used. (localhost is ignored in this case).
        """
        super().__init__(logger, multicast_port)
        self._my_ip_address = get_ip_for_adapter(adapterName)

        if self._my_ip_address is None:
            all_adapters = get_ipv4_ips()
            all_adapter_names = [ip.nice_name for ip in all_adapters]
            if forceAdapterName:
                raise RuntimeError(f'No adapter "{adapterName}" found. Having {all_adapter_names}')

            # see if there is only one physical adapter. if yes, use it
            adapters_not_localhost = [a for a in all_adapters if not a.ip.startswith('127.')]
            if len(adapters_not_localhost) == 1:
                self._my_ip_address = (adapters_not_localhost[0].ip,)  # a tuple
            else:
                raise RuntimeError(f'No adapter "{adapterName}" found. Having {all_adapter_names}')

    def _isAcceptedAddress(self, addr):
        """ check if any of the regular expressions matches the argument"""
        return addr in self._my_ip_address


_FallbackMedicalDeviceTypesFilter = [QName(NS_DPWS, 'Device'),
                                     QName('http://standards.ieee.org/downloads/11073/11073-20702-2016',
                                           'MedicalDevice')]



from sdc11073.roles import operationprovider
from . import alarmprovider
from . import audiopauseprovider
from . import clockprovider
from . import metricprovider
from . import contextprovider
from . import patientcontextprovider
from . import providerbase
from .. import loghelper
from .. import namespaces
from .. import sdcdevice


class GenericSetComponentStateOperationProvider(providerbase.ProviderRole):
    """
    Responsible for SetComponentState Operations
    """
    def makeOperationInstance(self, operationDescriptorContainer):
        """ Can handle following cases:
        SetComponentStateOperationDescriptor, target = any AbstractComponentDescriptor: => handler = _setComponentState
        """
        operationTargetHandle = operationDescriptorContainer.OperationTarget
        operationTargetDescriptorContainer = self._mdib.descriptions.handle.getOne(operationTargetHandle)

        if operationDescriptorContainer.NODETYPE == namespaces.domTag('SetComponentStateOperationDescriptor'):
            if operationTargetDescriptorContainer.NODETYPE in (namespaces.domTag('MdsDescriptor'),
                                                               namespaces.domTag('ChannelDescriptor'),
                                                               namespaces.domTag('VmdDescriptor'),
                                                               namespaces.domTag('ClockDescriptor'),
                                                               namespaces.domTag('ScoDescriptor'),
                                                               ):
                operation = self._mkOperation(sdcdevice.sco.SetComponentStateOperation,
                                              handle=operationDescriptorContainer.handle,
                                              operationTargetHandle=operationTargetHandle,
                                              codedValue=operationDescriptorContainer.Type,
                                              currentArgumentHandler=self._setComponentState)
                return operation
            return None # the operation target is no AbstractDeviceComponentDescriptor
        elif operationDescriptorContainer.NODETYPE == namespaces.domTag('ActivateOperationDescriptor'):
            #  on what can activate be called?
            if operationTargetDescriptorContainer.NODETYPE in (namespaces.domTag('MdsDescriptor'),
                                                               namespaces.domTag('ChannelDescriptor'),
                                                               namespaces.domTag('VmdDescriptor'),
                                                               namespaces.domTag('ScoDescriptor'),
                                                               ):
                # no generic handler to be called!
                return self._mkOperation(sdcdevice.sco.ActivateOperation,
                                              handle=operationDescriptorContainer.handle,
                                              operationTargetHandle=operationTargetHandle,
                                              codedValue=operationDescriptorContainer.Type)
            return None



    def _setComponentState(self, operationInstance, value):
        """

        :param operationInstance: the operation
        :param value: a list of proposed metric states
        :return:
        """
        #ToDo: consider ModifiableDate attribute
        operationInstance.currentValue = value
        with self._mdib.mdibUpdateTransaction() as mgr:
            for proposedComponentState in value:
                state = mgr.getComponentState(proposedComponentState.descriptorHandle)
                if state.isComponentState:
                    self._logger.info('updating {} with proposed component state', state)
                    state.updateFromOtherContainer(proposedComponentState, skippedProperties=['StateVersion', 'DescriptorVersion'])
                else:
                    self._logger.warn('_setComponentState operation: ignore invalid referenced type {} in operation', state.NODETYPE)


class BaseProduct(object):
    def __init__(self, log_prefix):
        # self.contextstate_provider = contextprovider.GenericContextProvider(log_prefix=log_prefix)  # default handler
        # self.componentstate_provider = GenericSetComponentStateOperationProvider(log_prefix=log_prefix) # default handler
        # self.metric_provider = None
        # self.patientcontext_provider = None
        # self.alarm_provider = None
        # self.alarm_provider = None
        # self.operation_provider = None
        # self.audiopause_provider = None
        # self.daynight_provider = None
        # self.clock_provider = None
        # self.ensembleContextProvider = None
        self._mdib = None
        self._ordered_providers = []    # order matters, each provider can hide operations of later ones
                                        # start with most specific providers, end with most general ones
        self._logger = loghelper.getLoggerAdapter('sdc.device.{}'.format(self.__class__.__name__), log_prefix)


    def _all_providers_sorted(self):
        return self._ordered_providers
        # # specialized roles first, generic roles last
        # return self._withoutNoneValues([self.audiopause_provider, self.daynight_provider, self.clock_provider,
        #         self.patientcontext_provider, self.alarm_provider,
        #         self.metric_provider, self.ensembleContextProvider, self.operation_provider,
        #         self.contextstate_provider,
        #         self.componentstate_provider])

    @staticmethod
    def _withoutNoneValues(some_list):
        return [ e for e in some_list if e is not None]


    def initOperations(self, mdib, sco):
        self._mdib = mdib
        """ register all actively provided operations """
        for role_handler in self._all_providers_sorted():
            role_handler.initOperations(mdib)

        self._registerExistingMdibOperations(sco)

        for role_handler in self._all_providers_sorted():
            ops = role_handler.makeMissingOperations()
            for op in ops:
                sco.registerOperation(op)

        # log all operations that do not have a handler now
        all_mdib_ops = []
        for nodetype in [namespaces.domTag('SetValueOperationDescriptorContainer'),
                         namespaces.domTag('SetStringOperationDescriptor'),
                         namespaces.domTag('SetContextStateOperationDescriptorContainer'),
                         namespaces.domTag('SetMetricStateOperationDescriptorContainer'),
                         namespaces.domTag('SetComponentStateOperationDescriptorContainer'),
                         namespaces.domTag('SetAlertStateOperationDescriptorContainer'),
                         namespaces.domTag('ActivateOperationDescriptorContainer')]:
            all_mdib_ops.extend(self._mdib.descriptions.NODETYPE.get(nodetype, []))
        all_mdib_op_handles = [op.Handle for op in all_mdib_ops]
        all_not_registered_op_handles = [op_h for op_h in all_mdib_op_handles if sco.getOperationByHandle(op_h) is None]
        if not all_mdib_op_handles:
            self._logger.info('this device has no operations in mdib.')
        elif all_not_registered_op_handles:
            self._logger.info('there are operations without handler! handles = {}', all_not_registered_op_handles)
        else:
            self._logger.info('there are no operations without handler.')
        mdib.mkStateContainersforAllDescriptors()
        mdib.preCommitHandler = self._onPreCommit
        mdib.postCommitHandler = self._onPostCommit

    def stop(self):
        for role_handler in self._all_providers_sorted():
            role_handler.stop()

    def makeOperationInstance(self, operationDescriptorContainer):
        """ try to get an operation for this operationDescriptorContainer ( given in mdib) """
        operationTargetHandle = operationDescriptorContainer.OperationTarget
        operationTargetDescr = self._mdib.descriptions.handle.getOne(operationTargetHandle, allowNone=True) # descriptor container
        if operationTargetDescr is None:
            # this operation is incomplete, the operation target does not exist. Registration not possible.
            self._logger.warn(
                'Operation {}: target {} does not exist, will not register operation'.format(operationDescriptorContainer.handle, operationTargetHandle))
            return
        for role_handler in self._all_providers_sorted():
            op = role_handler.makeOperationInstance(operationDescriptorContainer)
            if op is not None:
                self._logger.info('{} provided operation for {}'.format(role_handler.__class__.__name__, operationDescriptorContainer))
                return op
            else:
                self._logger.debug('{}: no handler for {}'.format(op.__class__.__name__, operationDescriptorContainer))
        return None

    def _registerExistingMdibOperations(self, sco):
        operationDescriptorContainers = self._mdib.getOperationDescriptors()
        for c in operationDescriptorContainers:
            registered_op = sco.getOperationByHandle(c.handle)
            if registered_op is None:
                self._logger.info('found unregistered {} in mdib, handle={}, code={} target={}'.format(
                    c.NODETYPE.localname, c.Handle, c.Type, c.OperationTarget))
                op = self.makeOperationInstance(c)
                if op is not None:
                    sco.registerOperation(op)

    def _onPreCommit(self, mdib, transaction):
        for p in self._all_providers_sorted():
            p.onPreCommit(mdib, transaction)
        self._addMissingStatesToTransaction(mdib, transaction)


    def _onPostCommit(self, mdib, transaction):
        for p in self._all_providers_sorted():
            p.onPostCommit(mdib, transaction)
        self._removeStatesForDeletedDescriptors(mdib, transaction)

    def _removeStatesForDeletedDescriptors(self, mdib, transaction):
        """
        remove states from mdib for deleted descriptors
        :return:
        """
        for tr_item in transaction.descriptorUpdates.values():
            if tr_item.new is None:
                # deleted descriptor
                objects = mdib.states.descriptorHandle.get(tr_item.old.handle)
                if objects:
                    mdib.states.removeObjects(objects)
                objects = mdib.contextStates.descriptorHandle.get(tr_item.old.handle)
                if objects:
                    mdib.contextStates.removeObjects(objects)


    def _addMissingStatesToTransaction(self, mdib, transaction):
        """
        add states to new descriptors if they are not part of this transaction
        """
        for tr_item in transaction.descriptorUpdates.values():
            if tr_item.old is None:
                # new descriptor
                state_cls = mdib.getStateClsForDescriptor(tr_item.new)
                if not state_cls.isMultiState:
                    if not transaction.hasState(tr_item.new.handle):
                        st = state_cls(mdib.nsmapper, tr_item.new)
                        st.updateNode()
                        transaction.addState(st)


class GenericProduct(BaseProduct):
    def __init__(self, audiopause_provider, daynight_provider, clock_provider, log_prefix):
        super().__init__(log_prefix)

        self._ordered_providers.extend([audiopause_provider, daynight_provider, clock_provider])
        self._ordered_providers.extend([patientcontextprovider.GenericPatientContextProvider(log_prefix=log_prefix),
                                       alarmprovider.GenericAlarmProvider(log_prefix=log_prefix),
                                       metricprovider.GenericMetricProvider(log_prefix=log_prefix),
                                       operationprovider.OperationProvider(log_prefix=log_prefix),
                                       GenericSetComponentStateOperationProvider(log_prefix=log_prefix)
        ])


class MinimalProduct(BaseProduct):
    def __init__(self, log_prefix=None):
        super().__init__(log_prefix)
        self.metric_provider = metricprovider.GenericMetricProvider(log_prefix=log_prefix) # needed in a test
        self._ordered_providers.extend([audiopauseprovider.GenericSDCAudioPauseProvider(log_prefix=log_prefix),
                                       clockprovider.GenericSDCClockProvider(log_prefix=log_prefix),
                                       patientcontextprovider.GenericPatientContextProvider(log_prefix=log_prefix),
                                       alarmprovider.GenericAlarmProvider(log_prefix=log_prefix),
                                       self.metric_provider,
                                       operationprovider.OperationProvider(log_prefix=log_prefix),
                                       GenericSetComponentStateOperationProvider(log_prefix=log_prefix)
        ])


class ExtendedProduct(BaseProduct):
    def __init__(self, log_prefix=None):
        super().__init__(log_prefix)
        self._ordered_providers.extend([audiopauseprovider.GenericSDCAudioPauseProvider(log_prefix=log_prefix),
                                       clockprovider.GenericSDCClockProvider(log_prefix=log_prefix),
                                       contextprovider.EnsembleContextProvider(log_prefix=log_prefix),
                                       contextprovider.LocationContextProvider(log_prefix=log_prefix),
                                       patientcontextprovider.GenericPatientContextProvider(log_prefix=log_prefix),
                                       alarmprovider.GenericAlarmProvider(log_prefix=log_prefix),
                                       self.metric_provider,
                                       operationprovider.OperationProvider(log_prefix=log_prefix),
                                       GenericSetComponentStateOperationProvider(log_prefix=log_prefix)
        ])



""" Using lxml based SoapClient"""
import functools
import logging
import os
import ssl
import traceback
import typing
from dataclasses import dataclass
from urllib.parse import urlparse, urlsplit

from lxml import etree as etree_

import sdc11073.certloader
from . import subscription
from .hostedservice import CTreeServiceClient, DescriptionEventClient, ContextServiceClient, WaveformClient
from .hostedservice import HostedServiceClient, GetServiceClient, SetServiceClient, StateEventClient
from .localizationservice import LocalizationServiceClient
from .operations import OperationsManager
from .. import commlog
from .. import compression
from .. import loghelper
from .. import netconn
from .. import observableproperties as properties
from ..definitions_base import ProtocolsRegistry
from ..definitions_sdc import SDC_v1_Definitions
from ..namespaces import Prefix_Namespace as Prefix
from ..namespaces import nsmap
from ..pysoap.soapclient import SoapClient
from ..pysoap.soapenvelope import GenericNode, WsAddress, Soap12Envelope, DPWSEnvelope, MetaDataSection

# default ssl context data
here = os.path.dirname(__file__)
caFolder = os.path.join(os.path.dirname(here), 'ca')
_ssl_certfile = os.path.join(caFolder, 'sdccert.pem')  # certification chain, contains root ca and signed public key
_ssl_keyfile = os.path.join(caFolder, 'userkey.pem')  # private key of own certificate
_ssl_cacert = os.path.join(caFolder, 'cacert.pem')  # common root ca that signed all sdc devices
_ssl_passwd = 'dummypass'  # 'Phase1' #dummypass
_ssl_cypherfile = os.path.join(caFolder, 'cyphers.json')  # Json file that determines ciphers to be used


class HostDescription(object):
    def __init__(self, dpws_envelope):
        self._dpws_envelope = dpws_envelope
        self.thisModel = dpws_envelope.thisModel
        self.thisDevice = dpws_envelope.thisDevice
        self.host = dpws_envelope.host

    def __str__(self):
        return 'HostDescription: thisModel = {}, thisDevice = {}, host = {}'.format(self.thisModel, self.thisDevice,
                                                                                    self.host)


class HostedServiceDescription(object):
    def __init__(self, service_id, endpoint_address, log_prefix=''):
        self._endpoint_address = endpoint_address
        self.service_id = service_id
        self.log_prefix = log_prefix
        self.metaData = None
        self.wsdl_string = None
        self.wsdl = None
        self._logger = loghelper.getLoggerAdapter('sdc.client.hosted', log_prefix)
        self._url = urlparse(endpoint_address)
        self.services = {}

    def readMetadata(self, soap_client):
        soapEnvelope = Soap12Envelope(nsmap)
        self._logger.debug('calling GetMetadata on {}', self._endpoint_address)
        soapEnvelope.setAddress(WsAddress(action='http://schemas.xmlsoap.org/ws/2004/09/mex/GetMetadata/Request',
                                          to=self._endpoint_address))
        soapEnvelope.addBodyObject(
            GenericNode(etree_.Element('{http://schemas.xmlsoap.org/ws/2004/09/mex}GetMetadata')))
        endpointEnvelope = soap_client.postSoapEnvelopeTo(self._url.path,
                                                          soapEnvelope,
                                                          msg='<{}> readMetadata'.format(self.service_id))
        self.metaData = MetaDataSection.fromEtreeNode(endpointEnvelope.bodyNode)
        self.readwsdl(soap_client, self.metaData.wsdl_location)
        return

    def readwsdl(self, soap_client, wsdl_url):
        p = urlparse(wsdl_url)
        actual_path = p.path + '?{}'.format(p.query) if p.query else p.path
        self.wsdl_string = soap_client.getUrl(actual_path, msg='{}:getwsdl'.format(self.log_prefix))
        commlog.defaultLogger.logWsdl(self.wsdl_string)
        try:
            self.wsdl = etree_.fromstring(self.wsdl_string,
                                          parser=etree_.ETCompatXMLParser())  # make am ElementTree instance
        except etree_.XMLSyntaxError as ex:
            self._logger.error(
                'could not read wsdl from {}: error={}, data=\n{}'.format(actual_path, ex, self.wsdl_string))

    def __repr__(self):
        return '{} "{}" endpoint = {}'.format(self.__class__.__name__, self.service_id, self._endpoint_address)


def ip2Int(ipString):
    """ Convert string like '192.168.0.1' to an integer
    (helper for sortIPAddresses)"""
    return functools.reduce(lambda x, y: x * 256 + y, (int(x) for x in ipString.split('.')), 0)


def _cmp(a, b, _refInt):
    """ helper for sortIPAddresses"""
    _a = abs(ip2Int(a) - _refInt)
    _b = abs(ip2Int(b) - _refInt)
    diff = _a - _b
    if diff < 0:
        return -1
    elif diff > 0:
        return 1
    else:
        return 0


def sortIPAddresses(adresses, refIp):
    """ sorts list addresses by distance to refIP, shortest distance first"""
    _ref = ip2Int(refIp)
    adresses.sort(key=lambda a: abs(ip2Int(a) - _ref))
    return adresses


@dataclass(frozen=True)
class SubscriptionEndData:
    subscription: subscription.ClSubscription
    soap_envelope: Soap12Envelope


class SdcClient(object):
    """ The SdcClient can be used with a known device location.
    The location is typically the result of a wsdiscovery process.
    This class expects that the BICEPS services are available in the device.
    What if not???? => raise exception in discoverHostedServices
    """
    isConnected = properties.ObservableProperty(False)  # a boolean

    # observable properties for all notifications
    # all incoming Notifications can be observed in stateEventReportEnvelope ( as soap envelope)
    stateEventReportEnvelope = properties.ObservableProperty()

    # the following observables can be used to observe the incoming notifications by message type.
    # They contain only the body node of the notification, not the envelope
    waveFormReport = properties.ObservableProperty()
    episodicMetricReport = properties.ObservableProperty()
    episodicAlertReport = properties.ObservableProperty()
    episodicComponentReport = properties.ObservableProperty()
    episodicOperationalStateReport = properties.ObservableProperty()
    episodicContextReport = properties.ObservableProperty()
    periodicMetricReport = properties.ObservableProperty()
    periodicAlertReport = properties.ObservableProperty()
    periodicComponentReport = properties.ObservableProperty()
    periodicOperationalStateReport = properties.ObservableProperty()
    periodicContextReport = properties.ObservableProperty()
    descriptionModificationReport = properties.ObservableProperty()
    operationInvokedReport = properties.ObservableProperty()
    subscriptionEndData = properties.ObservableProperty()  # SubscriptionEndData

    _servicesLookup = {'ContainmentTree': CTreeServiceClient,  # wpf naming
                       'ContainmentTreeService': CTreeServiceClient,  # sdc naming
                       'Get': GetServiceClient,
                       'GetService': GetServiceClient,
                       'StateEvent': StateEventClient,
                       'StateEventService': StateEventClient,
                       'Context': ContextServiceClient,
                       'ContextService': ContextServiceClient,
                       'Waveform': WaveformClient,
                       'WaveformService': WaveformClient,
                       'Set': SetServiceClient,
                       'SetService': SetServiceClient,
                       'DescriptionEvent': DescriptionEventClient,
                       'DescriptionEventService': DescriptionEventClient,
                       'LocalizationService': LocalizationServiceClient,
                       }

    SSL_CIPHERS = None  # None : use SSL default
    _ssl_context_container: typing.Optional[sdc11073.certloader.SSLContextContainer]

    def __init__(self, devicelocation, deviceType, validate=True, sslEvents='auto', sslContext=None,
                 my_ipaddress=None, logLevel=None, ident='',
                 soap_notifications_handler_class=None,
                 chunked_requests=False,
                 ssl_context_container: sdc11073.certloader.SSLContextContainer = None, ):  # pylint:disable=too-many-arguments
        """
        :param devicelocation: the XAddr location for meta data, e.g. http://10.52.219.67:62616/72c08f50-74cc-11e0-8092-027599143341
        :param deviceType: a QName that defines the device type, e.g. '{http://standards.ieee.org/downloads/11073/11073-20702-2016}MedicalDevice'
        :param sslEvents: define if client uses https
             sslEvents='auto': use https if Xaddress of device is https
             sslEvents=True: always use https
             sslEvents=False: use only http
        :param sslContext: the ssl context that shall be used for https connections. If None, https is not possible.
        :param my_ipaddress: This address is used for the http server that receives notifications.
             If value is None, best own address is determined automatically (recommended).
        :param ssl_context_container container containing a client and a server context
        """
        self._devicelocation = devicelocation
        self._soap_notifications_handler_class = soap_notifications_handler_class
        if deviceType is None:
            self.sdc_definitions = SDC_v1_Definitions
        else:
            self.sdc_definitions = None
            for definition_cls in ProtocolsRegistry.protocols:
                if definition_cls.ns_matches(deviceType.namespace):
                    self.sdc_definitions = definition_cls
                    break
            if self.sdc_definitions is None:
                raise ValueError('cannot create instance, no known BICEPS schema version identified')

        self._xml_validator = self.sdc_definitions.xml_validator
        splitted = urlsplit(self._devicelocation)
        self._device_uses_https = splitted.scheme.lower() == 'https'

        self.log_prefix = ident or ''
        self.chunked_requests = chunked_requests
        self._sslEvents = sslEvents
        self._setupLogging(logLevel)
        self._logger = loghelper.getLoggerAdapter('sdc.client', self.log_prefix)
        self._logger_wf = loghelper.getLoggerAdapter('sdc.client.wf', self.log_prefix)  # waveform logger
        if my_ipaddress is None:
            self._my_ipaddress = self._findBestOwnIpAddress()
        else:
            self._my_ipaddress = my_ipaddress
        self._logger.info('SdcClient for {} uses own IP Address {}', self._devicelocation, self._my_ipaddress)
        self.metaData = None
        self.hostDescription = None
        self._hostedServices = {}  # lookup by service id
        self._validate = validate

        try:
            self._logger.info('Using SSL is enabled. TLS 1.3 Support = {}', ssl.HAS_TLSv1_3)
        except AttributeError:
            self._logger.info('Using SSL is enabled. TLS 1.3 is not supported')

        if sslContext and ssl_context_container:
            raise ValueError('sslContext and ssl_context_container must not both be given')

        if sslContext:
            self._ssl_context_container = sdc11073.certloader.SSLContextContainer(client_context=sslContext,
                                                                                  server_context=sslContext)
        else:
            self._ssl_context_container = ssl_context_container

        self._notificationsDispatcherThread = None

        self._logger.info('created {} for {}', self.__class__.__name__, self._devicelocation)

        self._compression_methods = compression.encodings[:]
        self._subscriptionMgr = None
        self._operationsManager = None
        self._serviceClients = {}
        self._mdib = None
        self._soapClients = {}  # all http connections that this client holds
        self.peerCertificate = None
        self.binary_peer_cert = None
        self.all_subscribed = False

    def _register_mdib(self, mdib):
        """ SdcClient sometimes must know the mdib data (e.g. Set service, activate method)."""
        if mdib is not None and self._mdib is not None:
            raise RuntimeError('SdcClient has already an registered mdib')
        self._mdib = mdib
        if self.client('Set') is not None:
            self.client('Set').register_mdib(mdib)
        if self.client('Context') is not None:
            self.client('Context').register_mdib(mdib)

    @property
    def mdib(self):
        return self._mdib

    @property
    def my_ipaddress(self):
        return self._my_ipaddress

    def _findBestOwnIpAddress(self):
        myIpAddresses = netconn.get_ipv4_addresses()

        splitted = urlsplit(self._devicelocation)
        sortIPAddresses(myIpAddresses, splitted.hostname)
        return myIpAddresses[0]

    def _subscribe(self, dpwsHosted, actions, callback):
        """ creates a subscription object and registers it in
        :param dpwsHosted: proxy for the hosted service that provides the events we want to subscribe to
                           This is the target for all subscribe/unsubscribe ... messages
        :param actions: a list of filters. this (joined) string is sent to the sdc server in the Subscribe message
        :param callback: callable with signature callback(soapEnvlope)
        @return: a subscription object that has callback already registerd
        """
        s = self._subscriptionMgr.mkSubscription(dpwsHosted, actions)
        for f in actions:
            self._notificationsDispatcherThread.dispatcher.register_function(f,
                                                                             s.notifyTo_identifier.text,
                                                                             s.onNotification)

        # register callback for end of subscription
        self._notificationsDispatcherThread.dispatcher.register_function(self.sdc_definitions.Actions.SubscriptionEnd,
                                                                         s.end_to_identifier.text,
                                                                         self._onSubScriptionEnd)
        if callback is not None:
            properties.bind(s, notification=callback)
        s.subscribe()
        return s

    def client(self, porttypename):
        """ returns the client for the given port type name.
        WDP and SDC use different port type names, e.g. WPF="Get", SDC="GetService".
        If the port type is not found directly, it tries also with or without "Service" in name.
        :param porttypename: string, e.g "Get", or "GetService", ...
        """
        client = self._serviceClients.get(porttypename)
        if client is None and porttypename.endswith('Service'):
            client = self._serviceClients.get(porttypename[:-7])
        if client is None and not porttypename.endswith('Service'):
            client = self._serviceClients.get(porttypename + "Service")
        return client

    @property
    def GetService_client(self):
        return self.client('GetService')

    @property
    def SetService_client(self):
        return self.client('SetService')

    @property
    def DescriptionEventService_client(self):
        return self.client('DescriptionEventService')

    @property
    def StateEventService_client(self):
        return self.client('StateEventService')

    @property
    def ContextService_client(self):
        return self.client('ContextService')

    @property
    def WaveformService_client(self):
        return self.client('Waveform')

    @property
    def ContainmentTreeService_client(self):
        return self.client('ContainmentTreeService')

    @property
    def ArchiveService_client(self):
        return self.client('ArchiveService')

    @property
    def LocalizationService_client(self):
        return self.client('LocalizationService')

    @property
    def subscription_mgr(self):
        return self._subscriptionMgr

    def startAll(self, notSubscribedActions=None, subscriptionsCheckInterval=None, async_dispatch=True,
                 subscribe_periodic_reports=False, dispatcher_timeout=15.0):
        """
        :param notSubscribedActions: a list of pmtypes.Actions elements or None. if None, everything is subscribed.
        :param subscriptionsCheckInterval: an interval in seconds or None
        :param async_dispatch: if True, incoming requests are queued and response is sent immediately (processing is done later).
                                if False, response is sent after the complete processing is done.
        :param subscribe_periodic_reports: boolean
        :param dispatcher_timeout: time to wait for the event sink aka. dispatcher thread to be started, if timeout is
                                   exceeded a RuntimeError is raised
        @return: None
        """
        self.discoverHostedServices()
        self._startEventSink(async_dispatch, dispatcher_timeout=dispatcher_timeout)
        periodic_actions = {self.sdc_definitions.Actions.PeriodicMetricReport,
                            self.sdc_definitions.Actions.PeriodicAlertReport,
                            self.sdc_definitions.Actions.PeriodicComponentReport,
                            self.sdc_definitions.Actions.PeriodicContextReport,
                            self.sdc_definitions.Actions.PeriodicOperationalStateReport}
        # start subscription manager
        self._subscriptionMgr = subscription.SubscriptionManager(self._notificationsDispatcherThread.base_url,
                                                                 log_prefix=self.log_prefix,
                                                                 checkInterval=subscriptionsCheckInterval,
                                                                 xml_validator=self._xml_validator)
        self._subscriptionMgr.start()

        # flag 'self.all_subscribed' tells mdib that mdib state versions shall not have any gaps
        # => log warnings for missing versions
        self.all_subscribed = True
        notSubscribedActionsSet = set([])
        if notSubscribedActions:
            not_subscribed_episodic_actions = [a for a in notSubscribedActions if not 'Periodic' in a]
            if not_subscribed_episodic_actions:
                self.all_subscribed = False
                notSubscribedActionsSet = set(notSubscribedActions)

        # start operationInvoked subscription and tell all
        self._operationsManager = OperationsManager(self.log_prefix)

        for client in self._serviceClients.values():
            client.setOperationsManager(self._operationsManager)

        # start all subscriptions
        # group subscriptions per hosted service
        for service_id, dpwsHosted in self.metaData.hosted.items():
            available_actions = []
            if dpwsHosted.types:
                for port_type_qname in dpwsHosted.types:
                    port_type = port_type_qname.split(':')[-1]
                    client = self.client(port_type)
                    if client is not None:
                        available_actions.extend(client.getSubscribableActions())
            if len(available_actions) > 0:
                subscribe_actions = set(available_actions) - notSubscribedActionsSet
                if not subscribe_periodic_reports:
                    subscribe_actions -= set(periodic_actions)
                try:
                    self._subscribe(dpwsHosted, subscribe_actions,
                                    self._onAnyStateEventReport)
                except Exception as ex:
                    self.all_subscribed = False  # => do not log errors when mdib versions are missing in notifications
                    self._logger.error('startAll: could not subscribe: error = {}, actions= {}',
                                       traceback.format_exc(), subscribe_actions)

        # connect self.isConnected observable to allSubscriptionsOkay observable in subscriptionsmanager
        def setIsConnected(isOk):
            self.isConnected = isOk

        properties.strongbind(self._subscriptionMgr, allSubscriptionsOkay=setIsConnected)
        self.isConnected = self._subscriptionMgr.allSubscriptionsOkay

    def stopAll(self, unsubscribe=True, closeAllConnections=True):
        if self._subscriptionMgr is not None:
            if unsubscribe:
                self._subscriptionMgr.unsubscribeAll()
            self._subscriptionMgr.stop()
        self._stopEventSink(closeAllConnections)
        self._register_mdib(None)

        for cl in self._soapClients.values():
            cl.close()
        self._soapClients = {}

    def setUsedCompression(self, *compression_methods):
        # update list in place
        del self._compression_methods[:]
        self._compression_methods.extend(compression_methods)

    def getMetaData(self):
        _url = urlparse(self._devicelocation)
        wsc = self._getSoapClient(self._devicelocation)

        if wsc.isClosed():
            wsc.connect()

        if self._ssl_context_container is not None and _url.scheme == 'https':
            sock = wsc.sock
            self.peerCertificate = sock.getpeercert(binary_form=False)
            self.binary_peer_cert = sock.getpeercert(binary_form=True)

        soapEnvelope = Soap12Envelope(nsmap)
        soapEnvelope.setAddress(WsAddress(action='{}/Get'.format(Prefix.WXF.namespace),
                                          to=self._devicelocation))

        self.metaData = wsc.postSoapEnvelopeTo(_url.path, soapEnvelope, responseFactory=DPWSEnvelope.fromXMLString,
                                               msg='getMetadata')
        self.hostDescription = HostDescription(self.metaData)
        self._logger.debug('HostDescription: {}', self.hostDescription)

    def discoverHostedServices(self):
        """ Discovers all hosted services.
        Raises RuntimeError if device does not provide the expected BICEPS services
        """
        # we need to read the meta data of the device only once => temporary soap client is sufficient
        self._logger.debug('reading meta data from {}', self._devicelocation)
        if self.metaData is None:
            self.getMetaData()

        # now query also meta data of hosted services
        self._mkHostedServices()
        self._logger.debug('Services: {}', self._serviceClients.keys())

        # only GetService is mandatory!!!
        if self.GetService_client is None:
            raise RuntimeError('GetService not detected! found services = {}'.format(self._serviceClients.keys()))

    def _getSoapClient(self, address):
        _url = urlparse(address)
        key = (_url.scheme, _url.netloc)
        soap_client = self._soapClients.get(key)
        if soap_client is None:
            use_ssl = True
            if self._sslEvents == 'auto' and _url.scheme == 'http':
                use_ssl = False
            elif self._sslEvents is False:  # no ssl
                use_ssl = False
            else:  # only ssl
                if self._ssl_context_container is None:
                    raise RuntimeError('missing ssl context for ssl connection')
            soap_client = SoapClient(
                _url.netloc,
                logger=loghelper.getLoggerAdapter('sdc.client.soap', self.log_prefix),
                sslContext=(self._ssl_context_container.client_context
                            if use_ssl and self._ssl_context_container else None),
                sdc_definitions=self.sdc_definitions,
                supportedEncodings=self._compression_methods,
                requestEncodings=None,
                chunked_requests=self.chunked_requests,
                xml_validator=self._xml_validator if self._validate else None)

            self._soapClients[key] = soap_client
        return soap_client

    def _mkHostedServices(self):
        for hosted in self.metaData.hosted.values():
            endpoint_reference = hosted.endpointReferences[0].address
            soapClient = self._getSoapClient(endpoint_reference)
            hosted.soapClient = soapClient
            if hosted.types:
                ns_types = [t.split(':') for t in hosted.types]
            else:
                ns_types = []
            h_descr = HostedServiceDescription(hosted.serviceId, endpoint_reference, self.log_prefix)
            self._hostedServices[hosted.serviceId] = h_descr
            h_descr.readMetadata(soapClient)
            for _, porttype in ns_types:
                h = self._mkHostedServiceClient(porttype, soapClient, hosted)
                self._serviceClients[porttype] = h
                h_descr.services[porttype] = h

    def _mkHostedServiceClient(self, porttype, soapClient, hosted):
        cls = self._servicesLookup.get(porttype, HostedServiceClient)
        return cls(soapClient, hosted, porttype, self.sdc_definitions, self.log_prefix)

    def _startEventSink(self, async_dispatch, dispatcher_timeout=15.0):
        if self._sslEvents == 'auto':
            sslContext = (self._ssl_context_container.server_context
                          if self._device_uses_https and self._ssl_context_container else None)
        elif self._sslEvents:  # True
            sslContext = self._ssl_context_container.server_context if self._ssl_context_container else None
        else:  # False
            sslContext = None

        # create Event Server
        self._notificationsDispatcherThread = subscription.NotificationsReceiverDispatcherThread(
            self._my_ipaddress,
            sslContext,
            log_prefix=self.log_prefix,
            sdc_definitions=self.sdc_definitions,
            supportedEncodings=self._compression_methods,
            soap_notifications_handler_class=self._soap_notifications_handler_class,
            async_dispatch=async_dispatch)

        self._notificationsDispatcherThread.start()
        event_is_set = self._notificationsDispatcherThread.started_evt.wait(timeout=dispatcher_timeout)
        if not event_is_set:
            self._logger.error('Cannot start consumer, start event of EventSink not set.')
            raise RuntimeError('Cannot start consumer, start event of EventSink not set.')

        self._logger.info('serving EventSink on {}', self._notificationsDispatcherThread.base_url)

    def _stopEventSink(self, closeAllConnections):
        if self._notificationsDispatcherThread is not None:
            self._notificationsDispatcherThread.stop(closeAllConnections)

    def _onAnyStateEventReport(self, soapenvelope):
        """ dispatch by message body"""
        self.stateEventReportEnvelope = soapenvelope  # update observable
        message = soapenvelope.bodyNode[0].tag
        if message.endswith('EpisodicMetricReport'):
            return self._onEpisodicMetricReport(soapenvelope)
        elif message.endswith('EpisodicAlertReport'):
            return self._onEpisodicAlertReport(soapenvelope)
        elif message.endswith('EpisodicComponentReport'):
            return self._onEpisodicComponentReport(soapenvelope)
        elif message.endswith('EpisodicOperationalStateReport'):
            return self._onEpisodicOperationalStateReport(soapenvelope)
        elif message.endswith('EpisodicContextReport'):
            return self._onEpisodicContextReport(soapenvelope)
        elif message.endswith('WaveformStream') or message.endswith(
                'WaveformStreamReport'):  # different names in Draft6 and Final
            return self._onWaveFormReport(soapenvelope)
        elif message.endswith('OperationInvokedReport'):
            return self._onOperationInvokedReport(soapenvelope)
        elif message.endswith('EpisodicContextReport'):
            return self._onEpisodicContextReport(soapenvelope)
        elif message.endswith('DescriptionModificationReport'):
            return self._onDescriptionReport(soapenvelope)
        elif message.endswith('PeriodicMetricReport'):
            return self._onPeriodicMetricReport(soapenvelope)
        elif message.endswith('PeriodicAlertReport'):
            return self._onPeriodicAlertReport(soapenvelope)
        elif message.endswith('PeriodicComponentReport'):
            return self._onPeriodicComponentReport(soapenvelope)
        elif message.endswith('PeriodicOperationalStateReport'):
            return self._onPeriodicOperationalStateReport(soapenvelope)
        elif message.endswith('PeriodicContextReport'):
            return self._onPeriodicContextReport(soapenvelope)
        else:
            raise RuntimeError('unknown message {}'.format(message))

    def _onOperationInvokedReport(self, soapenvelope):
        ret = self._operationsManager.onOperationInvokedReport(soapenvelope)
        report = soapenvelope.bodyNode.xpath('msg:OperationInvokedReport', namespaces=nsmap)
        self.operationInvokedReport = report[0]  # update observable
        return ret

    def _onWaveFormReport(self, soapenvelope):
        try:
            waveformStream = soapenvelope.bodyNode[0]  # the msg:WaveformStreamReport node
        except IndexError:
            waveformStream = None

        if waveformStream is not None:
            self._logger_wf.debug('_onWaveFormReport')
        else:
            self._logger_wf.error('WaveformStream does not contain msg:WaveformStream!', soapenvelope)

        self.waveFormReport = waveformStream  # update observable

    def _get_report(self, soap_envelope, name):
        reports = soap_envelope.bodyNode.xpath(f'msg:{name}', namespaces=nsmap)
        if len(reports) == 1:
            self._logger.debug('_get_report {}', name)
            return reports[0]
        elif len(reports) > 1:
            self._logger.error('report contains {} elements of msg:{}!', len(reports), name)
        else:
            self._logger.error('report does not contain msg:{}!', name)

    def _onEpisodicMetricReport(self, soapenvelope):
        report = self._get_report(soapenvelope, 'EpisodicMetricReport')
        if report is not None:
            self.episodicMetricReport = report

    def _onPeriodicMetricReport(self, soapenvelope):
        report = self._get_report(soapenvelope, 'PeriodicMetricReport')
        if report is not None:
            self.periodicMetricReport = report

    def _onEpisodicAlertReport(self, soapenvelope):
        report = self._get_report(soapenvelope, 'EpisodicAlertReport')
        if report is not None:
            self.episodicAlertReport = report

    def _onPeriodicAlertReport(self, soapenvelope):
        report = self._get_report(soapenvelope, 'PeriodicAlertReport')
        if report is not None:
            self.periodicAlertReport = report

    def _onEpisodicComponentReport(self, soapenvelope):
        report = self._get_report(soapenvelope, 'EpisodicComponentReport')
        if report is not None:
            self.episodicComponentReport = report

    def _onPeriodicComponentReport(self, soapenvelope):
        report = self._get_report(soapenvelope, 'PeriodicComponentReport')
        if report is not None:
            self.periodicComponentReport = report

    def _onEpisodicOperationalStateReport(self, soapenvelope):
        report = self._get_report(soapenvelope, 'EpisodicOperationalStateReport')
        if report is not None:
            self.episodicOperationalStateReport = report

    def _onPeriodicOperationalStateReport(self, soapenvelope):
        report = self._get_report(soapenvelope, 'PeriodicOperationalStateReport')
        if report is not None:
            self.periodicOperationalStateReport = report

    def _onEpisodicContextReport(self, soapenvelope):
        report = self._get_report(soapenvelope, 'EpisodicContextReport')
        if report is not None:
            self.episodicContextReport = report

    def _onPeriodicContextReport(self, soapenvelope):
        report = self._get_report(soapenvelope, 'PeriodicContextReport')
        if report is not None:
            self.periodicContextReport = report

    def _onDescriptionReport(self, soapenvelope):
        report = self._get_report(soapenvelope, 'DescriptionModificationReport')
        if report is not None:
            self.descriptionModificationReport = report

    def _onSubScriptionEnd(self, soapenvelope):
        self.stateEventReportEnvelope = soapenvelope  # update observable
        the_subscription = self._subscriptionMgr.onSubScriptionEnd(soapenvelope)  # subscription can be None
        # write info to observable
        self.subscriptionEndData = SubscriptionEndData(the_subscription, soapenvelope)

    def _setupLogging(self, logLevel):
        loghelper.ensureLogStream()
        if logLevel is None:
            return
        clientLog = logging.getLogger('sdc.client')
        clientLog.setLevel(logLevel)

    def __str__(self):
        return 'SdcClient to {} {} on {}'.format(self.hostDescription.thisDevice,
                                                 self.hostDescription.thisModel,
                                                 self._devicelocation)

    @classmethod
    def fromWsdService(cls, wsdService, validate=True, sslEvents='auto',
                       sslContext=None, ssl_context_container: sdc11073.certloader.SSLContextContainer = None,
                       my_ipaddress=None, logLevel=logging.INFO,
                       ident='', soap_notifications_handler_class=None):
        device_locations = wsdService.getXAddrs()
        if not device_locations:
            raise RuntimeError('discovered Service has no address!{}'.format(wsdService))
        device_location = device_locations[0]
        deviceType = None
        for _qname in wsdService.getTypes():
            qname = etree_.QName(_qname.namespace, _qname.localname)
            for protocol in ProtocolsRegistry.protocols:
                if protocol.ns_matches(qname):
                    deviceType = protocol.MedicalDeviceType
                    break
        return cls(device_location, deviceType=deviceType, validate=validate, sslEvents=sslEvents,
                   sslContext=sslContext, ssl_context_container=ssl_context_container,
                   my_ipaddress=my_ipaddress, logLevel=logLevel, ident=ident,
                   soap_notifications_handler_class=soap_notifications_handler_class)



import traceback
import time
from threading import Lock
from dataclasses import dataclass
from lxml import etree as etree_
from .. import observableproperties as properties
from .. import namespaces
from .. import pmtypes
from .. import multikey
from typing import Union

class RtSampleContainer(object):
    """Contains a single Value"""
    def __init__(self, valueString, timestamp, validity, annotations = None):
        self.valueString = valueString
        self.value = float(valueString)
        self.observationTime = timestamp
        self.validity = validity
        self.annotations = [] if annotations is None else annotations
    
    @property
    def age(self):
        return time.time() - self.observationTime
    
    def __repr__(self):
        return 'RtSample value="{}" validity="{}" time={}'.format(self.valueString, self.validity, self.observationTime)


@dataclass
class MdibVersionGroup:
    mdib_version: int
    sequence_id: str
    instance_id: Union[int, None]

    def update_node(self, node):
        """sets attributes in Node"""
        node.set('MdibVersion', str(self.mdib_version))
        node.set('SequenceId', self.sequence_id)
        if self.instance_id is not None:
            node.set('InstanceId', str(self.instance_id))


class _MultikeyWithVersionLookup(multikey.MultiKeyLookup):
    """
    This class keeps track of versions of removed objects
    """
    def __init__(self):
        multikey.MultiKeyLookup.__init__(self)
        self.handle_version_lookup = dict()


    def removeObject(self, obj):
        if obj is not None:
            self._saveVersion(obj)
        multikey.MultiKeyLookup.removeObject(self, obj)


    def removeObjectNoLock(self, obj):
        if obj is not None:
            self._saveVersion(obj)
        multikey.MultiKeyLookup.removeObjectNoLock(self, obj)

    def removeObjectsNoLock(self, objs):
        for obj in objs:
            if obj is not None:
                self._saveVersion(obj)
        multikey.MultiKeyLookup.removeObjectsNoLock(self, objs)


class DescriptorsLookup(_MultikeyWithVersionLookup):
    """ This class knows about the hierarchy of descriptors and keeps the order of objects """

    def __init__(self):
        _MultikeyWithVersionLookup.__init__(self)
        self.addIndex('handle', multikey.UIndexDefinition(lambda obj: obj.handle))
        self.addIndex('parentHandle', multikey.IndexDefinition(lambda obj: obj.parentHandle))
        # nodeName may differ but refers to the same type. use NODETYPE instead
        self.addIndex('nodeName', multikey.IndexDefinition(lambda obj: obj.nodeName))
        self.addIndex('NODETYPE', multikey.IndexDefinition(lambda obj: obj.NODETYPE))
        self.addIndex('ConditionSignaled', multikey.IndexDefinition(lambda obj: obj.ConditionSignaled, indexNoneValues=False))
        # an index to find all alert conditions for a metric (AlertCondition is the only class that has a
        # "Source" attribute, therefore this simple approach without type testing is sufficient):
        self.addIndex('Source', multikey.IndexDefinition1n(lambda obj: [s.text for s in obj.Source], indexNoneValues=False))

        # ToDo: really 3 indices for coding needed?
        self.addIndex('codingSystem', multikey.IndexDefinition(lambda obj: obj.codingSystem))
        self.addIndex('codeId', multikey.IndexDefinition(lambda obj: obj.codeId))
        self.addIndex('coding', multikey.IndexDefinition(lambda obj: obj.coding))


    def _saveVersion(self, obj):
        self.handle_version_lookup[obj.handle] = obj.DescriptorVersion

    def setVersion(self, obj, increment=True):
        version = self.handle_version_lookup.get(obj.handle)
        if version is not None:
            if increment:
                version +=1
            obj.DescriptorVersion = version

    def addObject(self, obj):
        with self._lock:
            self.addObjectNoLock(obj)

    def addObjectNoLock(self, obj):
        """ appends obj to parent"""
        _MultikeyWithVersionLookup.addObjectNoLock(self, obj)
        parent = None if obj.parentHandle is None else self.handle.getOne(obj.parentHandle, allowNone=True)
        if parent is not None:
            parent.addChild(obj)

    def addObjects(self, objs):
        with self._lock:
            self.addObjectsNoLock(objs)

    def addObjectsNoLock(self, objs):
        for obj in objs:
            self.addObjectNoLock(obj)

    def removeObject(self, obj):
        keys = self._objectIDs.get(id(obj))
        if keys is None:
            return
        with self._lock:
            self.removeObjectNoLock(obj)

    def removeObjectNoLock(self, obj):
        _MultikeyWithVersionLookup.removeObjectNoLock(self, obj)
        parent = self.handle.getOne(obj.parentHandle, allowNone=True)
        if parent is not None:
            parent.rmChild(obj)

    def removeObjects(self, objs):
        with self._lock:
            self.removeObjectsNoLock(objs)

    def removeObjectsNoLock(self, objs):
        for obj in objs:
            self.removeObjectNoLock(obj)

    def replaceObject(self, newObj):
        with self._lock:
            self.replaceObjectNoLock(newObj)

    def replaceObjectNoLock(self, newObj):
        """ remove existing descriptorContainer and add new one, but do not touch childlist of parent (that keeps order)"""
        origObj = self.handle.getOne(newObj.handle)
        self.removeObjectNoLock(origObj)
        self.addObjectNoLock(newObj)


class StatesLookup(_MultikeyWithVersionLookup):
    def _saveVersion(self, obj):
        self.handle_version_lookup[obj.descriptorHandle] = obj.StateVersion

    def setVersion(self, obj, increment=True):
        version = self.handle_version_lookup.get(obj.descriptorHandle)
        if version is not None:
            if increment:
                version +=1
            obj.StateVersion = version


class MultiStatesLookup(_MultikeyWithVersionLookup):
    def _saveVersion(self, obj):
        self.handle_version_lookup[obj.Handle] = obj.StateVersion

    def setVersion(self, obj, increment=True):
        version = self.handle_version_lookup.get(obj.Handle)
        if version is not None:
            if increment:
                version +=1
            obj.StateVersion = version


class MdibContainer(object):

    # these observables can be used to watch any change of data in the mdib. They contain lists of containers that were changed.
    # every transaction (devicemdib) or notification (client mdib) will report their changes here.
    metricsByHandle = properties.ObservableProperty(fireOnlyOnChangedValue=False)
    waveformByHandle = properties.ObservableProperty(fireOnlyOnChangedValue=False)
    alertByHandle = properties.ObservableProperty(fireOnlyOnChangedValue=False)
    contextByHandle = properties.ObservableProperty(fireOnlyOnChangedValue=False)
    componentByHandle = properties.ObservableProperty(fireOnlyOnChangedValue=False)
    newDescriptorByHandle = properties.ObservableProperty(fireOnlyOnChangedValue=False)
    updatedDescriptorByHandle = properties.ObservableProperty(fireOnlyOnChangedValue=False)
    deletedDescriptorByHandle = properties.ObservableProperty(fireOnlyOnChangedValue=False)
    operationByHandle = properties.ObservableProperty(fireOnlyOnChangedValue=False)
    deletedStatesByHandle = properties.ObservableProperty(fireOnlyOnChangedValue=False) # is a result of deleted descriptors
    sequenceId = properties.ObservableProperty()
    instanceId = properties.ObservableProperty()

    def __init__(self, sdc_definitions):
        """
        :param sdc_definitions: a class derived from Definitions_Base
        """
        self.sdc_definitions = sdc_definitions
        self._logger = None # must to be instantiated by derived class
        self.nsmapper = namespaces.DocNamespaceHelper()  # default map, might be replaced with nsmap from xml file  
        self.mdibVersion = 0
        self.sequenceId = ''  # needs to be set to a reasonable value by derived class
        self.instanceId = None # None or an unsigned int
        self.log_prefix = ''
        
        self.descriptions = DescriptorsLookup()

        self.states = StatesLookup() #multikey.MultiKeyLookup()
        self.states.addIndex('descriptorHandle', multikey.UIndexDefinition(lambda obj: obj.descriptorHandle))
        self.states.addIndex('NODETYPE', multikey.IndexDefinition(lambda obj: obj.NODETYPE, indexNoneValues=False))

        self.contextStates = MultiStatesLookup() #multikey.MultiKeyLookup()

        # descriptorHandle index is NOT unique!
        # => multiple ContextStates refer to the same descriptor( history of locations)
        # 'handle' index can be unique, because we ignore None values 
        self.contextStates.addIndex('descriptorHandle', multikey.IndexDefinition(lambda obj: obj.descriptorHandle))
        self.contextStates.addIndex('handle', multikey.UIndexDefinition(lambda obj: obj.Handle, indexNoneValues=False))
        self.contextStates.addIndex('NODETYPE', multikey.IndexDefinition(lambda obj: obj.NODETYPE, indexNoneValues=False))
        self.mdibLock = Lock()
        

        self.mdStateVersion = 0
        self.mdDescriptionVersion = 0

    @property
    def logger(self):
        return self._logger

    @property
    def mdib_version_group(self):
        return MdibVersionGroup(self.mdibVersion, self.sequenceId, self.instanceId)

    def addDescriptionContainers(self, descriptionContainers):
        """ init self.descriptions with provided descriptors
        :param descriptionContainers: a list od DescriptionStateContainer objects
        """
        newDescriptorByHandle = {}
        with self.descriptions._lock: #pylint: disable=protected-access
            for d in descriptionContainers:
                self.descriptions.addObjectNoLock(d)
                newDescriptorByHandle[d.handle] = d

        # finally update observable property
        if newDescriptorByHandle:
            self.newDescriptorByHandle = newDescriptorByHandle


    def clearStates(self):
        """removes all states and context states. """
        with self.states._lock: #pylint: disable=protected-access
            self.states.clear()
            self.contextStates.clear()

        # clear also the observable properties
        self.metricsByHandle = None
        self.waveformByHandle = None
        self.alertByHandle = None
        self.contextByHandle = None
        self.componentByHandle = None
        self.operationByHandle = None


    def _updateStateObservables(self, statecontainer_list):
        metricsByHandle = {}
        waveformByHandle = {}
        alertByHandle = {}
        contextByHandle = {}
        componentByHandle = {}
        operationByHandle = {}
        for sc in statecontainer_list:
            # add state to the corresponding dictionary, depending on type
            if sc.isAlertState:
                alertByHandle[sc.descriptorHandle] = sc
            elif sc.isRealtimeSampleArrayMetricState:  # test for this class before AbstractMetricStateContainer!!
                waveformByHandle[sc.descriptorHandle] = sc
            elif sc.isMetricState:
                metricsByHandle[sc.descriptorHandle] = sc
            elif sc.isComponentState:
                componentByHandle[sc.descriptorHandle] = sc
            elif sc.isOperationalState:
                operationByHandle[sc.descriptorHandle] = sc
            elif sc.isContextState:
                contextByHandle[sc.Handle] = sc
            elif sc.isSystemContextState or sc.isMultiState:
                pass   # ignoring for now
            elif sc.NODETYPE == namespaces.domTag('ScoState'): # special case Draft6 ScoState (is not a component state)
                pass  # this cannot be updated anyway over the network, but handle it here to avoid runtime error
            else:
                raise RuntimeError('handling of {} has been forgotten to implement!'.format(sc.__class__.__name__))

        #finally update observable properties
        if alertByHandle:
            self.alertByHandle = alertByHandle
        if waveformByHandle:
            self.waveformByHandle = waveformByHandle
        if metricsByHandle:
            self.metricsByHandle = metricsByHandle
        if componentByHandle:
            self.componentByHandle = componentByHandle
        if operationByHandle:
            self.operationByHandle = operationByHandle
        if contextByHandle:
            self.contextByHandle = contextByHandle


    def addStateContainers(self, stateContainers):
        """Adds states to self.states and self.contextStates.
        :param stateContainers: a list of StateContainer objects.
        """
        for sc in stateContainers:
            if sc.descriptorContainer is not None:
                self._logger.debug('addStateContainers: new state {}', sc)
            else:
                self._logger.warn('addStateContainers: new state {}, but has no descriptor!', sc)

            my_multikey = self.contextStates if sc.isContextState else self.states
            try:
                my_multikey.addObject(sc)
            except KeyError as ex:
                self._logger.error('addStateContainers: {}, keys={}; {}', ex,
                                   my_multikey.Handle.keys(), traceback.format_exc())

        # finally update observable properties
        self._updateStateObservables(stateContainers)

    setMdStates = addStateContainers # backwards compatibility


    def _reconstructMdDescription(self):
        """build dom tree from current data
        @return: an etree_ node
        """
        doc_nsmap = self.nsmapper.docNssmap
        rootContainers = self.descriptions.parentHandle.get(None) or []
        mdDescriptionNode = etree_.Element(namespaces.domTag('MdDescription'),
                                           attrib={'DescriptionVersion':str(self.mdDescriptionVersion)},
                                           nsmap=doc_nsmap)

        def connectDescriptors(parentContainer, parentNode):
            childContainers = parentContainer.getOrderedChildContainers()
            ret = parentContainer.connectChildContainers(parentNode, childContainers)
            # recursive call for children
            for childContainer, node in ret:
                connectDescriptors(childContainer, node)

        for rootContainer in rootContainers:
            n = rootContainer.mkDescriptorNode()
            mdDescriptionNode.append(n)
            connectDescriptors(rootContainer, n)
        return mdDescriptionNode


    def _reconstructMdib(self, addContextStates):
        """build dom tree from current data
        :param addContextStates: bool
        @return: an etree_ node
        """
        doc_nsmap = self.nsmapper.docNssmap
        mdibNode = etree_.Element(namespaces.msgTag('Mdib'), nsmap=doc_nsmap)
        self.mdib_version_group.update_node(mdibNode)
        mdDescriptionNode = self._reconstructMdDescription()
        mdibNode.append(mdDescriptionNode)

        # add a list of states
        mdStateNode = etree_.SubElement(mdibNode, namespaces.domTag('MdState'),
                                        attrib={'StateVersion':str(self.mdStateVersion)},
                                        nsmap=doc_nsmap)
        for stateContainer in self.states.objects:
            try:
                tmpNode = stateContainer.mkStateNode()
                mdStateNode.append(tmpNode)
            except RuntimeError:
                self._logger.error('State {} has no descriptorContainer', stateContainer.descriptorHandle)
        if addContextStates:
            for stateContainer in self.contextStates.objects:
                tmpNode = stateContainer.mkStateNode()
                mdStateNode.append(tmpNode)

        return mdibNode

    def reconstructMdDescription(self):
        """build dom tree from current data
        @return: a tuple etree_ node, mdibVersion
        """
        with self.mdibLock:
            node = self._reconstructMdDescription()
            return node, self.mdib_version_group

    def reconstructMdib(self):
        """build dom tree from current data
        This method does not include context states!
        @return: an etree_ node
        """
        with self.mdibLock:
            return self._reconstructMdib(addContextStates=False), self.mdib_version_group


    def reconstructMdibWithContextStates(self):
        """ this method includes the context states in mdib tree.
        """
        with self.mdibLock:
            return self._reconstructMdib(addContextStates=True), self.mdib_version_group


    def nodeToString(self, etree_node, pretty_print=False, xml_declaration=True, encoding='utf-8'):
        """Special toString converter that replaces the internal normalized namespaces with the correct external namespaces.
        @return: a string
        """
        mdibString = etree_.tostring(etree_node, pretty_print=pretty_print, xml_declaration=xml_declaration, encoding=encoding)
        return self.sdc_definitions.denormalizeXMLText(mdibString)


    def get_children_by_type_match(self, parent_handle: str, coding: Union[pmtypes.CodedValue, pmtypes.Coding]):
        """
        Returns all descriptors with matching parentHandle and Type.
        """
        all_children = self.descriptions.parentHandle.get(parent_handle)
        return [child for child in all_children if pmtypes.have_matching_codes(child.Type, coding)]

    def getDescriptorByCode(self, vmdCode: Union[pmtypes.CodedValue, pmtypes.Coding],
                            channelCode: Union[pmtypes.CodedValue, pmtypes.Coding],
                            metricCode: Union[pmtypes.CodedValue, pmtypes.Coding]):
        """ Find a descriptor by its Type hierarchy.
        :param vmdCode: a pmtypes.CodedValue or a pmtypes.Coding instance
        :param channelCode: a pmtypes.CodedValue or a pmtypes.Coding instance
        :param metricCode: a pmtypes.CodedValue or a pmtypes.Coding instance
        :return: None or a DescriptorContainer
        """
        all_vmds = self.descriptions.NODETYPE.get(namespaces.domTag('VmdDescriptor'))
        matching_vmd_handles = [vmd.Handle for vmd in all_vmds if pmtypes.have_matching_codes(vmd.Type, vmdCode)]

        matching_channels = []
        for handle in matching_vmd_handles:
            matching_channels.extend(self.get_children_by_type_match(handle, channelCode ))
        matching_channel_handles = [ch.Handle for ch in matching_channels]

        matching_leaf_descriptors = []
        for handle in matching_channel_handles:
            matching_leaf_descriptors.extend(self.get_children_by_type_match(handle, metricCode ))
        if len(matching_leaf_descriptors) == 0:
            return
        if len(matching_leaf_descriptors) > 1:
            raise RuntimeError('found multiple channel descriptors for vmd={} channel={} metric={}'.format(vmdCode, channelCode, metricCode))
        return matching_leaf_descriptors[0]

    getMetricDescriptorByCode = getDescriptorByCode  # backwards compatibility with previous name

    def getOperationsForMetric(self, vmdCode, channelCode, metricCode):
        """ This is the "correct" way to find an operation.
        Using well known handles is shaky, because they have no meaning and can change over time!
        :param vmdCode: a pmtypes.CodedValue or a pmtypes.Coding instance
        :param channelCode: a pmtypes.CodedValue or a pmtypes.Coding instance
        :param metricCode: a pmtypes.CodedValue or a pmtypes.Coding instance
        @return: a list of matching Operation Containers
        """
        descriptorContainer = self.getDescriptorByCode(vmdCode, channelCode, metricCode)
        return self.getOperationDescriptorsForDescriptorHandle(descriptorContainer.handle)


    def getOperationDescriptorsForDescriptorHandle(self, descriptorHandle, **additionalFilters):
        """
        :param descriptorHandle: the handle for that operations shall be found
        :return: a list with operation descriptors that have descriptorHandle as OperationTarget. List can be empty
        :additionalFilters: optional filters for the key = name of member attribute, value = expected value
            example: NODETYPE=domTag('SetContextStateOperationDescriptor') filters for SetContextStateOperation descriptors
        """
        allOperationContainers = self.getOperationDescriptors()
        myOperations = [opC for opC in allOperationContainers if opC.OperationTarget == descriptorHandle]
        for k, v in additionalFilters.items():
            myOperations = [op for op in myOperations if getattr(op, k) == v]
        return myOperations


    def getStateContainerClass(self, qNameType):
        """
        :param qNameType: a QName instance
        """
        cls = self.sdc_definitions.sc.getContainerClass(qNameType)
        if cls is None:
            self._logger.warn('No class for type={}; using AbstractStateContainer', str(qNameType))
            cls = self.sdc_definitions.sc.AbstractStateContainer
        return cls

    def getStateClsForDescriptor(self, descriptorContainer):
        stateClassQType = descriptorContainer.STATE_QNAME
        if stateClassQType is None:
            raise TypeError('No state association for {}'.format(descriptorContainer.__class__.__name__))
        return self.getStateContainerClass(stateClassQType)


    def mkStateContainerFromDescriptor(self, descriptorContainer):
        cls = self.getStateClsForDescriptor(descriptorContainer)
        if cls is None:
            raise TypeError('No state container class for descr={}, name={}, type={}'.format(descriptorContainer.__class__.__name__, descriptorContainer.nodeName, descriptorContainer.nodeType))
        return cls(self.nsmapper, descriptorContainer)


    def getOperationDescriptors(self):
        """
        :return: a list of all operation descriptors
        """
        result = []
        for nodeType in ('SetValueOperationDescriptor',
                         'SetStringOperationDescriptor',
                         'ActivateOperationDescriptor',
                         'SetContextStateOperationDescriptor',
                         'SetMetricStateOperationDescriptor',
                         'SetComponentStateOperationDescriptor',
                         'SetAlertStateOperationDescriptor'):
            result.extend(self.descriptions.NODETYPE.get(namespaces.domTag(nodeType), []))
        return result


    def getDescriptorContainerClass(self, qNameType):
        """
        :param qNameType: a QName instance
        """
        cls = self.sdc_definitions.dc.getContainerClass(qNameType)
        if cls is None:
            self._logger.warn('No class for type={}; using AbstractDescriptorContainer',
                              str(qNameType))
            raise RuntimeError('No class for type={}; using AbstractDescriptorContainer'.format(str(qNameType)))
        return cls


    def selectDescriptors(self, *codings):
        """ Returns all descriptor containers that match a path defined by list of codings.
        example: 
        ['70041'] returns all containers that have CodedValue = 70041
        ['70041', '69650'] : returns all descriptors with CodedValue= 69650 and parent descriptor CodedValue = 70041
        ['70041', '69650', '69651'] : returns all descriptors with CodedValue= 69651 and parent descriptor CodedValue = 69650 and parent's parent descriptor CodedValue = 70041
        It is not necessary that path starts at the top of an mds, it can start anywhere.  
        """
        selectedObjects = None
        for coding in codings:
            
            if selectedObjects is None:
                selectedObjects = self.descriptions.objects # initially all objects
            else:
                # get all children of selected objects
                allhandles = [o.handle for o in selectedObjects]
                selectedObjects = []
                for h in allhandles:
                    selectedObjects.extend(self.descriptions.parentHandle.get(h, []))
                
            # normalize coding
            if isinstance(coding, str):
                coding = pmtypes.CodedValue(coding, pmtypes.DefaultCodingSystem).coding
            elif hasattr(coding, 'coding'):
                coding=coding.coding
                
            if coding is not None:
                # apply filter
                tmpObjects = [o for o in selectedObjects if o.coding == coding ]
                selectedObjects = tmpObjects
        return selectedObjects                    


    def getAllDescriptorsInSubTree(self, descriptorContainer, depthFirst=True, includeRoot=True):
        """ walks the tree below descriptorContainer.
        :param descriptorContainer:
        :param depthFirst: determines order of returned list. DepthFirst=True has all leaves on top, otherwise at the end.
        :param includeRoot: if True descriptorContainer itself is also part of returned list
        :return: a list of DescriptorContainer objects
        """
        result = []
        def _getchildren(parent):
            childContainers = self.descriptions.parentHandle.get(parent.handle, list())
            if not depthFirst:
                result.extend(childContainers)
            for ch in childContainers:
                _getchildren(ch)
            if depthFirst:
                result.extend(childContainers)
        if includeRoot and not depthFirst:
            result.append(descriptorContainer)
        _getchildren(descriptorContainer)
        if includeRoot and depthFirst:
            result.append(descriptorContainer)
        return result


    def _rmDescriptorsAndStates(self, descriptorContainers):
        """ recursive delete of a descriptor and all children and all related states"""
        deletedDescriptorByHandle = {}
        deletedStatesByHandle = {}
        for descriptorContainer in descriptorContainers:
            self._logger.debug('rm Descriptor node {} handle {}',
                               descriptorContainer.nodeName, descriptorContainer.handle)
            self.descriptions.removeObject(descriptorContainer)
            deletedDescriptorByHandle[descriptorContainer.handle] = descriptorContainer
            for m_key in (self.states, self.contextStates):
                stateContainers = m_key.descriptorHandle.get(descriptorContainer.handle)
                if stateContainers is not None:
                    # make a copy, otherwise removeObjects will manipulate same list in place
                    stateContainers = stateContainers[:]
                    self._logger.debug('rm {} states(s) associated to descriptor {} ',
                                      len(stateContainers), descriptorContainer.handle)
                    m_key.removeObjects(stateContainers)
                    deletedStatesByHandle[descriptorContainer.handle] = stateContainers

        if deletedDescriptorByHandle:
            self.deletedDescriptorByHandle = deletedDescriptorByHandle
        if deletedStatesByHandle:
            self.deletedStatesByHandle = deletedStatesByHandle


    def rmDescriptorHandleAll(self, handle):
        descriptorContainer = self.descriptions.handle.getOne(handle, allowNone=True)
        if descriptorContainer is not None:
            allDescriptors = self.getAllDescriptorsInSubTree(descriptorContainer)
            self._rmDescriptorsAndStates(allDescriptors)



from lxml import etree as etree_

from sdc11073 import namespaces, xmlparsing, pmtypes


class MdibStructureError(Exception):
    pass


class MessageReader(object):
    """ This class does all the conversions from DOM trees (body of SOAP messages) to MDIB objects."""
    def __init__(self, mdib):
        self._mdib = mdib
        self._logger = mdib.logger
        self._logPrefix = ''


    @staticmethod
    def getMdibRootNode(sdc_definitions, xml_text):
        """
        Creates a normalized and validated elementtree from xml_text.
        normalizing means that draft6 or final BICEPS namespaces are replaced by an standardized internal namespace.
        :param xml_text: xml document
        :return: elementtree node of the root element
        """
        xml_text = sdc_definitions.normalizeXMLText(xml_text)
        parser = etree_.ETCompatXMLParser(remove_comments=True, remove_blank_text=True)
        root = etree_.fromstring(xml_text, parser=parser, base_url=None)
        if root.tag != namespaces.msgTag('GetMdibResponse'):
            getmdibResponseNodes = root.xpath('//msg:GetMdibResponse', namespaces=namespaces.nsmap)
            if len(getmdibResponseNodes) != 1:
                raise ValueError('provided dom does not contain a msg:GetMdibResponse node!')
            else:
                root = getmdibResponseNodes[0]
        return root


    def readMdDescription(self, node):
        """
        Parses a GetMdDescriptionResponse or the MdDescription part of GetMdibResponse
        :param node: An etree node
        :return: a list of DescriptorContainer objects, sorted depth last
        """
        descriptions = []
        mdDescriptionNodes = node.xpath('//dom:MdDescription', namespaces=namespaces.nsmap)
        if not mdDescriptionNodes:
            raise ValueError('no MdDescription node found in tree')
        else:
            mdDescriptionNode = mdDescriptionNodes[0]

        def addChildren(parentNode):
            p_handle = parentNode.get('Handle')
            for childNode in parentNode:
                if childNode.get('Handle') is not None:
                    container = self.mkDescriptorContainerFromNode(childNode, p_handle)
                    descriptions.append(container)
                    addChildren(childNode)

        # iterate over tree, collect all handles of vmds, channels and metric descriptors
        allmds = mdDescriptionNode.findall(namespaces.domTag('Mds'))
        for mdsNode in allmds:
            mds = self.mkDescriptorContainerFromNode(mdsNode, None)
            descriptions.append(mds)
            addChildren(mdsNode)
        return descriptions


    def readMdState(self, node, additionalDescriptorContainers=None):
        """
        Parses a GetMdStateResponse or the MdState part of GetMdibResponse
        :param node: A node that contains MdState nodes
        :param additionalDescriptorContainers: a list of descriptor containers that can also be used for state creation
                (typically used if descriptors and states are created in the same transaction. In that case the descriptors are not yet part of mdib.)
        :return: a list of state containers
        """
        stateContainers = []
        mdStateNodes = node.xpath('//dom:MdState', namespaces=namespaces.nsmap)
        if mdStateNodes:
            allstates = mdStateNodes[0].findall(namespaces.domTag('State'))
            for state in allstates:
                try:
                    stateContainers.append(self.mkStateContainerFromNode(state, additionalDescriptorContainers=additionalDescriptorContainers))
                except MdibStructureError as ex:
                    self._logger.error('{}readMdState: cannot create: {}', self._logPrefix, ex)
        return stateContainers


    def readContextState(self, getContextStatesResponseNode):
        """ Creates Context State Containers from dom tree.
        :param getContextstatesResponseNode: node "getContextStatesResponse" of getContextStates.
        :param additionalDescriptorContainers: a list of descriptor containers that can also be used for state creation
                (typically used if descriptors and states are created in the same transaction. In that case the descriptors are not yet part of mdib.)
        @return: a list of state containers
        """
        states = []
        contextStateNodes = list(getContextStatesResponseNode) # list of msg:ContextStatenodes
        for contextStateNode in contextStateNodes:
            # hard remame to dom:State
            contextStateNode.tag = namespaces.domTag('State')
            try:
                stateContainer = self.mkStateContainerFromNode(contextStateNode)
                states.append(stateContainer)
            except MdibStructureError as ex:
                self._logger.error('{}readContextState: cannot create: {}', self._logPrefix, ex)
        return states


    def mkDescriptorContainerFromNode(self, node, parentHandle):
        """

        :param node: a descriptor node
        :param parentHandle: the handle of the parent
        :return: a DescriptorContainer object representing the content of node
        """
        nodeType = node.get(namespaces.QN_TYPE)
        if nodeType is not None:
            nodeType = namespaces.txt2QName(nodeType, node.nsmap)
        else:
            nodeType = etree_.QName(node.tag)
        cls = self._mdib.getDescriptorContainerClass(nodeType)
        return cls.fromNode(self._mdib.nsmapper, node, parentHandle)


    def mkStateContainerFromNode(self, node, forcedType=None, additionalDescriptorContainers = None):
        """
        :param node: a etree node
        :param forcedType: if given, the QName that shall be used for class instantiation instead of the data in node
        """
        if forcedType is not None:
            nodeType = forcedType
        else:
            nodeType = node.get(namespaces.QN_TYPE)
            if nodeType is not None:
                nodeType = namespaces.txt2QName(nodeType, node.nsmap)

        descriptorHandle = node.get('DescriptorHandle')
        descriptorContainer = self._mdib.descriptions.handle.getOne(descriptorHandle, allowNone=True)
        if descriptorContainer is None:
            if additionalDescriptorContainers is not None:
                correspondingDescriptors = [ d for d in additionalDescriptorContainers if d.handle == descriptorHandle]
            else:
                correspondingDescriptors = None
            if correspondingDescriptors is None or len(correspondingDescriptors) == 0:
                raise MdibStructureError(
                    'new state {}: descriptor with handle "{}" does not exist!'.format(nodeType.localname,
                                                                                       descriptorHandle))
            else:
                descriptorContainer = correspondingDescriptors[0]
        cls = self._mdib.getStateContainerClass(nodeType)
        if node.tag != namespaces.domTag('State'):
            node = xmlparsing.copy_node(node)  # make a copy, do not modify the original report
            node.tag = namespaces.domTag('State')
        return cls(self._mdib.nsmapper, descriptorContainer, node)


    def _mkStateContainersFromReportPart(self, reportPartNode):
        containers = []
        for childNode in reportPartNode.xpath('./*[not(self::msg:SourceMds)]', namespaces=namespaces.nsmap):
            desc_h = childNode.get('DescriptorHandle')
            if desc_h is None:
                self._logger.error('{}_onEpisodicComponentReport: missing descriptor handle in {}!',
                                   self._logPrefix, lambda:etree_.tostring(childNode))  #pylint: disable=cell-var-from-loop
            else:
                containers.append(self.mkStateContainerFromNode(childNode))
        return containers


    def readWaveformReport(self, reportNode):
        """
        Parses a waveform report
        :param reportNode: A waveform report etree
        :return: a list of StateContainer objects
        """
        states = []
        allSampleArrays = list(reportNode)
        for sampleArray in allSampleArrays:
            if sampleArray.tag.endswith('State'): # ignore everything else, e.g. Extension
                sc = self.mkStateContainerFromNode(sampleArray, namespaces.domTag('RealTimeSampleArrayMetricState'))
                states.append(sc)
        return states


    def readEpisodicMetricReport(self, reportNode):
        """
        Parses an episodic metric report
        :param reportNode:  An episodic metric report etree
        :return: a list of StateContainer objects
        """
        states = []
        reportPartNodes = reportNode.xpath('msg:ReportPart', namespaces=namespaces.nsmap)
        for reportPartNode in reportPartNodes:
            states.extend(self._mkStateContainersFromReportPart(reportPartNode))
        return states


    def readEpisodicAlertReport(self, reportNode):
        """
        Parses an episodic alert report
        :param reportNode:  An episodic alert report etree
        :return: a list of StateContainer objects
        """
        states = []
        allAlerts = reportNode.xpath('msg:ReportPart/msg:AlertState', namespaces=namespaces.nsmap)
        for alert in allAlerts:
            sc = self.mkStateContainerFromNode(alert)
            states.append(sc)
        return states


    def readOperationalStateReport(self, reportNode):
        """
        Parses an operational state report
        :param reportNode:  An operational state report etree
        :return: a list of StateContainer objects
        """
        states = []
        allOperationStateNodes = reportNode.xpath('msg:ReportPart/msg:OperationState', namespaces=namespaces.nsmap)
        for opStateNode in allOperationStateNodes:
            sc = self.mkStateContainerFromNode(opStateNode)
            states.append(sc)
        return states


    def readEpisodicContextReport(self, reportNode):
        """
        Parses an episodic context report
        :param reportNode:  An episodic context report etree
        :return: a list of StateContainer objects
        """
        states = []
        reportPartNodes = reportNode.xpath('msg:ReportPart', namespaces=namespaces.nsmap)
        for reportPartNode in reportPartNodes:
            sc = self._mkStateContainersFromReportPart(reportPartNode)
            states.extend(sc)
        return states


    def readEpisodicComponentReport(self, reportNode):
        """
        Parses an episodic component report
        :param reportNode:  An episodic component report etree
        :return: a list of StateContainer objects
        """
        states = []
        componentStateNodes = reportNode.xpath('msg:ReportPart/msg:ComponentState', namespaces=namespaces.nsmap)
        for componentState in componentStateNodes:
            sc = self.mkStateContainerFromNode(componentState)
            states.append(sc)
        return states


    def readDescriptionModificationReport(self, reportNode):
        """
        Parses a description modification report
        :param reportNode:  A description modification report etree
        :return: a list of DescriptorContainer objects
        """
        descriptors_list = []
        reportParts = list(reportNode) # list of msg:ReportPart nodes
        for reportPart in reportParts:
            descriptors = {pmtypes.DescriptionModificationTypes.UPDATE: ([], []),
                           pmtypes.DescriptionModificationTypes.CREATE: ([], []),
                           pmtypes.DescriptionModificationTypes.DELETE: ([], []),
                           }
            descriptors_list.append(descriptors)
            parentDescriptor = reportPart.get('ParentDescriptor')
            modificationType = reportPart.get('ModificationType', 'Upt')  # implied Value is 'Upt'
            descriptorNodes = reportPart.findall(namespaces.msgTag('Descriptor'))
            for descriptorNode in descriptorNodes:
                dc = self.mkDescriptorContainerFromNode(descriptorNode, parentDescriptor)
                descriptors[modificationType][0].append(dc)
            stateNodes = reportPart.findall(namespaces.msgTag('State'))
            for stateNode in stateNodes:
                sc = self.mkStateContainerFromNode(stateNode, additionalDescriptorContainers=descriptors[modificationType][0])
                descriptors[modificationType][1].append(sc)
        return descriptors_list
